import pytest
import salt.states.win_snmp as win_snmp
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {win_snmp: {}}


def test_agent_settings():
    """
    Test - Manage the SNMP sysContact, sysLocation, and sysServices settings.
    """
    kwargs = {
        "name": "agent-settings",
        "contact": "TestContact",
        "location": "TestLocation",
        "services": ["Internet"],
    }
    ret = {
        "name": kwargs["name"],
        "changes": {},
        "comment": "Agent settings already contain the provided values.",
        "result": True,
    }
    # Using this instead of dictionary comprehension in order to make pylint happy.
    get_ret = {key: value for (key, value) in kwargs.items() if key != "name"}
    mock_value_get = MagicMock(return_value=get_ret)
    mock_value_set = MagicMock(return_value=True)
    with patch.dict(
        win_snmp.__salt__,
        {
            "win_snmp.get_agent_settings": mock_value_get,
            "win_snmp.set_agent_settings": mock_value_set,
        },
    ):
        with patch.dict(win_snmp.__opts__, {"test": False}):
            assert win_snmp.agent_settings(**kwargs) == ret


def test_auth_traps_enabled():
    """
    Test - Manage the sending of authentication traps.
    """
    kwargs = {"name": "auth-traps", "status": True}
    ret = {
        "name": kwargs["name"],
        "changes": {"old": False, "new": True},
        "comment": "Set EnableAuthenticationTraps to contain the provided value.",
        "result": True,
    }
    mock_value_get = MagicMock(return_value=False)
    mock_value_set = MagicMock(return_value=True)
    with patch.dict(
        win_snmp.__salt__,
        {
            "win_snmp.get_auth_traps_enabled": mock_value_get,
            "win_snmp.set_auth_traps_enabled": mock_value_set,
        },
    ):
        with patch.dict(win_snmp.__opts__, {"test": False}):
            assert win_snmp.auth_traps_enabled(**kwargs) == ret
        with patch.dict(win_snmp.__opts__, {"test": True}):
            ret["comment"] = "EnableAuthenticationTraps will be changed."
            ret["result"] = None
            assert win_snmp.auth_traps_enabled(**kwargs) == ret


def test_community_names():
    """
    Test - Manage the SNMP accepted community names and their permissions.
    """
    kwargs = {
        "name": "community-names",
        "communities": {"TestCommunity": "Read Create"},
    }
    ret = {
        "name": kwargs["name"],
        "changes": {},
        "comment": "Communities already contain the provided values.",
        "result": True,
    }
    mock_value_get = MagicMock(return_value=kwargs["communities"])
    mock_value_set = MagicMock(return_value=True)
    with patch.dict(
        win_snmp.__salt__,
        {
            "win_snmp.get_community_names": mock_value_get,
            "win_snmp.set_community_names": mock_value_set,
        },
    ):
        with patch.dict(win_snmp.__opts__, {"test": False}):
            assert win_snmp.community_names(**kwargs) == ret



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""


import time

import salt.modules.ldapmod as ldapmod
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class LdapmodTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.ldapmod
    """

    def setup_loader_modules(self):
        return {ldapmod: {}}

    # 'search' function tests: 1

    def test_search(self):
        """
        Test if it run an arbitrary LDAP query and return the results.
        """

        class MockConnect:
            """
            Mocking _connect method
            """

            def __init__(self):
                self.bdn = None
                self.scope = None
                self._filter = None
                self.attrs = None

            def search_s(self, bdn, scope, _filter, attrs):
                """
                Mock function for search_s
                """
                self.bdn = bdn
                self.scope = scope
                self._filter = _filter
                self.attrs = attrs
                return "SALT"

        mock = MagicMock(return_value=True)
        with patch.dict(ldapmod.__salt__, {"config.option": mock}):
            with patch.object(
                ldapmod, "_connect", MagicMock(return_value=MockConnect())
            ):
                with patch.object(time, "time", MagicMock(return_value=8e-04)):
                    self.assertDictEqual(
                        ldapmod.search(filter="myhost"),
                        {
                            "count": 4,
                            "results": "SALT",
                            "time": {"raw": "0.0", "human": "0.0ms"},
                        },
                    )



"""
Azure (ARM) Resource State Module

.. versionadded:: 2019.2.0

:maintainer: <devops@decisionlab.io>
:maturity: new
:depends:
    * `azure <https://pypi.python.org/pypi/azure>`_ >= 2.0.0
    * `azure-common <https://pypi.python.org/pypi/azure-common>`_ >= 1.1.8
    * `azure-mgmt <https://pypi.python.org/pypi/azure-mgmt>`_ >= 1.0.0
    * `azure-mgmt-compute <https://pypi.python.org/pypi/azure-mgmt-compute>`_ >= 1.0.0
    * `azure-mgmt-network <https://pypi.python.org/pypi/azure-mgmt-network>`_ >= 1.7.1
    * `azure-mgmt-resource <https://pypi.python.org/pypi/azure-mgmt-resource>`_ >= 1.1.0
    * `azure-mgmt-storage <https://pypi.python.org/pypi/azure-mgmt-storage>`_ >= 1.0.0
    * `azure-mgmt-web <https://pypi.python.org/pypi/azure-mgmt-web>`_ >= 0.32.0
    * `azure-storage <https://pypi.python.org/pypi/azure-storage>`_ >= 0.34.3
    * `msrestazure <https://pypi.python.org/pypi/msrestazure>`_ >= 0.4.21
:platform: linux

:configuration: This module requires Azure Resource Manager credentials to be passed as a dictionary of
    keyword arguments to the ``connection_auth`` parameter in order to work properly. Since the authentication
    parameters are sensitive, it's recommended to pass them to the states via pillar.

    Required provider parameters:

    if using username and password:
      * ``subscription_id``
      * ``username``
      * ``password``

    if using a service principal:
      * ``subscription_id``
      * ``tenant``
      * ``client_id``
      * ``secret``

    Optional provider parameters:

    **cloud_environment**: Used to point the cloud driver to different API endpoints, such as Azure GovCloud. Possible values:
      * ``AZURE_PUBLIC_CLOUD`` (default)
      * ``AZURE_CHINA_CLOUD``
      * ``AZURE_US_GOV_CLOUD``
      * ``AZURE_GERMAN_CLOUD``

    Example Pillar for Azure Resource Manager authentication:

    .. code-block:: yaml

        azurearm:
            user_pass_auth:
                subscription_id: 3287abc8-f98a-c678-3bde-326766fd3617
                username: fletch
                password: 123pass
            mysubscription:
                subscription_id: 3287abc8-f98a-c678-3bde-326766fd3617
                tenant: ABCDEFAB-1234-ABCD-1234-ABCDEFABCDEF
                client_id: ABCDEFAB-1234-ABCD-1234-ABCDEFABCDEF
                secret: XXXXXXXXXXXXXXXXXXXXXXXX
                cloud_environment: AZURE_PUBLIC_CLOUD

    Example states using Azure Resource Manager authentication:

    .. code-block:: jinja

        {% set profile = salt['pillar.get']('azurearm:mysubscription') %}
        Ensure resource group exists:
            azurearm_resource.resource_group_present:
                - name: my_rg
                - location: westus
                - tags:
                    how_awesome: very
                    contact_name: Elmer Fudd Gantry
                - connection_auth: {{ profile }}

        Ensure resource group is absent:
            azurearm_resource.resource_group_absent:
                - name: other_rg
                - connection_auth: {{ profile }}

"""


import json
import logging

import salt.utils.files

__virtualname__ = "azurearm_resource"

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only make this state available if the azurearm_resource module is available.
    """
    if "azurearm_resource.resource_group_check_existence" in __salt__:
        return __virtualname__
    return (False, "azurearm_resource module could not be loaded")


def resource_group_present(
    name, location, managed_by=None, tags=None, connection_auth=None, **kwargs
):
    """
    .. versionadded:: 2019.2.0

    Ensure a resource group exists.

    :param name:
        Name of the resource group.

    :param location:
        The Azure location in which to create the resource group. This value cannot be updated once
        the resource group is created.

    :param managed_by:
        The ID of the resource that manages this resource group. This value cannot be updated once
        the resource group is created.

    :param tags:
        A dictionary of strings can be passed as tag metadata to the resource group object.

    :param connection_auth:
        A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.

    Example usage:

    .. code-block:: yaml

        Ensure resource group exists:
            azurearm_resource.resource_group_present:
                - name: group1
                - location: eastus
                - tags:
                    contact_name: Elmer Fudd Gantry
                - connection_auth: {{ profile }}

    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        ret[
            "comment"
        ] = "Connection information must be specified via connection_auth dictionary!"
        return ret

    group = {}

    present = __salt__["azurearm_resource.resource_group_check_existence"](
        name, **connection_auth
    )

    if present:
        group = __salt__["azurearm_resource.resource_group_get"](
            name, **connection_auth
        )
        ret["changes"] = __utils__["dictdiffer.deep_diff"](
            group.get("tags", {}), tags or {}
        )

        if not ret["changes"]:
            ret["result"] = True
            ret["comment"] = "Resource group {} is already present.".format(name)
            return ret

        if __opts__["test"]:
            ret["comment"] = "Resource group {} tags would be updated.".format(name)
            ret["result"] = None
            ret["changes"] = {"old": group.get("tags", {}), "new": tags}
            return ret

    elif __opts__["test"]:
        ret["comment"] = "Resource group {} would be created.".format(name)
        ret["result"] = None
        ret["changes"] = {
            "old": {},
            "new": {
                "name": name,
                "location": location,
                "managed_by": managed_by,
                "tags": tags,
            },
        }
        return ret

    group_kwargs = kwargs.copy()
    group_kwargs.update(connection_auth)

    group = __salt__["azurearm_resource.resource_group_create_or_update"](
        name, location, managed_by=managed_by, tags=tags, **group_kwargs
    )
    present = __salt__["azurearm_resource.resource_group_check_existence"](
        name, **connection_auth
    )

    if present:
        ret["result"] = True
        ret["comment"] = "Resource group {} has been created.".format(name)
        ret["changes"] = {"old": {}, "new": group}
        return ret

    ret["comment"] = "Failed to create resource group {}! ({})".format(
        name, group.get("error")
    )
    return ret


def resource_group_absent(name, connection_auth=None):
    """
    .. versionadded:: 2019.2.0

    Ensure a resource group does not exist in the current subscription.

    :param name:
        Name of the resource group.

    :param connection_auth:
        A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.
    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        ret[
            "comment"
        ] = "Connection information must be specified via connection_auth dictionary!"
        return ret

    group = {}

    present = __salt__["azurearm_resource.resource_group_check_existence"](
        name, **connection_auth
    )

    if not present:
        ret["result"] = True
        ret["comment"] = "Resource group {} is already absent.".format(name)
        return ret

    elif __opts__["test"]:
        group = __salt__["azurearm_resource.resource_group_get"](
            name, **connection_auth
        )

        ret["comment"] = "Resource group {} would be deleted.".format(name)
        ret["result"] = None
        ret["changes"] = {
            "old": group,
            "new": {},
        }
        return ret

    group = __salt__["azurearm_resource.resource_group_get"](name, **connection_auth)
    deleted = __salt__["azurearm_resource.resource_group_delete"](
        name, **connection_auth
    )

    if deleted:
        present = False
    else:
        present = __salt__["azurearm_resource.resource_group_check_existence"](
            name, **connection_auth
        )

    if not present:
        ret["result"] = True
        ret["comment"] = "Resource group {} has been deleted.".format(name)
        ret["changes"] = {"old": group, "new": {}}
        return ret

    ret["comment"] = "Failed to delete resource group {}!".format(name)
    return ret


def policy_definition_present(
    name,
    policy_rule=None,
    policy_type=None,
    mode=None,
    display_name=None,
    description=None,
    metadata=None,
    parameters=None,
    policy_rule_json=None,
    policy_rule_file=None,
    template="jinja",
    source_hash=None,
    source_hash_name=None,
    skip_verify=False,
    connection_auth=None,
    **kwargs
):
    """
    .. versionadded:: 2019.2.0

    Ensure a security policy definition exists.

    :param name:
        Name of the policy definition.

    :param policy_rule:
        A YAML dictionary defining the policy rule. See `Azure Policy Definition documentation
        <https://docs.microsoft.com/en-us/azure/azure-policy/policy-definition#policy-rule>`_ for details on the
        structure. One of ``policy_rule``, ``policy_rule_json``, or ``policy_rule_file`` is required, in that order of
        precedence for use if multiple parameters are used.

    :param policy_rule_json:
        A text field defining the entirety of a policy definition in JSON. See `Azure Policy Definition documentation
        <https://docs.microsoft.com/en-us/azure/azure-policy/policy-definition#policy-rule>`_ for details on the
        structure. One of ``policy_rule``, ``policy_rule_json``, or ``policy_rule_file`` is required, in that order of
        precedence for use if multiple parameters are used. Note that the `name` field in the JSON will override the
        ``name`` parameter in the state.

    :param policy_rule_file:
        The source of a JSON file defining the entirety of a policy definition. See `Azure Policy Definition
        documentation <https://docs.microsoft.com/en-us/azure/azure-policy/policy-definition#policy-rule>`_ for
        details on the structure. One of ``policy_rule``, ``policy_rule_json``, or ``policy_rule_file`` is required,
        in that order of precedence for use if multiple parameters are used. Note that the `name` field in the JSON
        will override the ``name`` parameter in the state.

    :param skip_verify:
        Used for the ``policy_rule_file`` parameter. If ``True``, hash verification of remote file sources
        (``http://``, ``https://``, ``ftp://``) will be skipped, and the ``source_hash`` argument will be ignored.

    :param source_hash:
        This can be a source hash string or the URI of a file that contains source hash strings.

    :param source_hash_name:
        When ``source_hash`` refers to a hash file, Salt will try to find the correct hash by matching the
        filename/URI associated with that hash.

    :param policy_type:
        The type of policy definition. Possible values are NotSpecified, BuiltIn, and Custom. Only used with the
        ``policy_rule`` parameter.

    :param mode:
        The policy definition mode. Possible values are NotSpecified, Indexed, and All. Only used with the
        ``policy_rule`` parameter.

    :param display_name:
        The display name of the policy definition. Only used with the ``policy_rule`` parameter.

    :param description:
        The policy definition description. Only used with the ``policy_rule`` parameter.

    :param metadata:
        The policy definition metadata defined as a dictionary. Only used with the ``policy_rule`` parameter.

    :param parameters:
        Required dictionary if a parameter is used in the policy rule. Only used with the ``policy_rule`` parameter.

    :param connection_auth:
        A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.

    Example usage:

    .. code-block:: yaml

        Ensure policy definition exists:
            azurearm_resource.policy_definition_present:
                - name: testpolicy
                - display_name: Test Policy
                - description: Test policy for testing policies.
                - policy_rule:
                    if:
                      allOf:
                        - equals: Microsoft.Compute/virtualMachines/write
                          source: action
                        - field: location
                          in:
                            - eastus
                            - eastus2
                            - centralus
                    then:
                      effect: deny
                - connection_auth: {{ profile }}

    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        ret[
            "comment"
        ] = "Connection information must be specified via connection_auth dictionary!"
        return ret

    if not policy_rule and not policy_rule_json and not policy_rule_file:
        ret["comment"] = (
            'One of "policy_rule", "policy_rule_json", or "policy_rule_file" is'
            " required!"
        )
        return ret

    if (
        sum(x is not None for x in [policy_rule, policy_rule_json, policy_rule_file])
        > 1
    ):
        ret["comment"] = (
            'Only one of "policy_rule", "policy_rule_json", or "policy_rule_file" is'
            " allowed!"
        )
        return ret

    if (policy_rule_json or policy_rule_file) and (
        policy_type or mode or display_name or description or metadata or parameters
    ):
        ret["comment"] = (
            'Policy definitions cannot be passed when "policy_rule_json" or'
            ' "policy_rule_file" is defined!'
        )
        return ret

    temp_rule = {}
    if policy_rule_json:
        try:
            temp_rule = json.loads(policy_rule_json)
        except Exception as exc:  # pylint: disable=broad-except
            ret["comment"] = "Unable to load policy rule json! ({})".format(exc)
            return ret
    elif policy_rule_file:
        try:
            # pylint: disable=unused-variable
            sfn, source_sum, comment_ = __salt__["file.get_managed"](
                None,
                template,
                policy_rule_file,
                source_hash,
                source_hash_name,
                None,
                None,
                None,
                __env__,
                None,
                None,
                skip_verify=skip_verify,
                **kwargs
            )
        except Exception as exc:  # pylint: disable=broad-except
            ret["comment"] = 'Unable to locate policy rule file "{}"! ({})'.format(
                policy_rule_file, exc
            )
            return ret

        if not sfn:
            ret["comment"] = 'Unable to locate policy rule file "{}"!)'.format(
                policy_rule_file
            )
            return ret

        try:
            with salt.utils.files.fopen(sfn, "r") as prf:
                temp_rule = json.load(prf)
        except Exception as exc:  # pylint: disable=broad-except
            ret["comment"] = 'Unable to load policy rule file "{}"! ({})'.format(
                policy_rule_file, exc
            )
            return ret

        if sfn:
            salt.utils.files.remove(sfn)

    policy_name = name
    if policy_rule_json or policy_rule_file:
        if temp_rule.get("name"):
            policy_name = temp_rule.get("name")
        policy_rule = temp_rule.get("properties", {}).get("policyRule")
        policy_type = temp_rule.get("properties", {}).get("policyType")
        mode = temp_rule.get("properties", {}).get("mode")
        display_name = temp_rule.get("properties", {}).get("displayName")
        description = temp_rule.get("properties", {}).get("description")
        metadata = temp_rule.get("properties", {}).get("metadata")
        parameters = temp_rule.get("properties", {}).get("parameters")

    policy = __salt__["azurearm_resource.policy_definition_get"](
        name, azurearm_log_level="info", **connection_auth
    )

    if "error" not in policy:
        if policy_type and policy_type.lower() != policy.get("policy_type", "").lower():
            ret["changes"]["policy_type"] = {
                "old": policy.get("policy_type"),
                "new": policy_type,
            }

        if (mode or "").lower() != policy.get("mode", "").lower():
            ret["changes"]["mode"] = {"old": policy.get("mode"), "new": mode}

        if (display_name or "").lower() != policy.get("display_name", "").lower():
            ret["changes"]["display_name"] = {
                "old": policy.get("display_name"),
                "new": display_name,
            }

        if (description or "").lower() != policy.get("description", "").lower():
            ret["changes"]["description"] = {
                "old": policy.get("description"),
                "new": description,
            }

        rule_changes = __utils__["dictdiffer.deep_diff"](
            policy.get("policy_rule", {}), policy_rule or {}
        )
        if rule_changes:
            ret["changes"]["policy_rule"] = rule_changes

        meta_changes = __utils__["dictdiffer.deep_diff"](
            policy.get("metadata", {}), metadata or {}
        )
        if meta_changes:
            ret["changes"]["metadata"] = meta_changes

        param_changes = __utils__["dictdiffer.deep_diff"](
            policy.get("parameters", {}), parameters or {}
        )
        if param_changes:
            ret["changes"]["parameters"] = param_changes

        if not ret["changes"]:
            ret["result"] = True
            ret["comment"] = "Policy definition {} is already present.".format(name)
            return ret

        if __opts__["test"]:
            ret["comment"] = "Policy definition {} would be updated.".format(name)
            ret["result"] = None
            return ret

    else:
        ret["changes"] = {
            "old": {},
            "new": {
                "name": policy_name,
                "policy_type": policy_type,
                "mode": mode,
                "display_name": display_name,
                "description": description,
                "metadata": metadata,
                "parameters": parameters,
                "policy_rule": policy_rule,
            },
        }

    if __opts__["test"]:
        ret["comment"] = "Policy definition {} would be created.".format(name)
        ret["result"] = None
        return ret

    # Convert OrderedDict to dict
    if isinstance(metadata, dict):
        metadata = json.loads(json.dumps(metadata))
    if isinstance(parameters, dict):
        parameters = json.loads(json.dumps(parameters))

    policy_kwargs = kwargs.copy()
    policy_kwargs.update(connection_auth)

    policy = __salt__["azurearm_resource.policy_definition_create_or_update"](
        name=policy_name,
        policy_rule=policy_rule,
        policy_type=policy_type,
        mode=mode,
        display_name=display_name,
        description=description,
        metadata=metadata,
        parameters=parameters,
        **policy_kwargs
    )

    if "error" not in policy:
        ret["result"] = True
        ret["comment"] = "Policy definition {} has been created.".format(name)
        return ret

    ret["comment"] = "Failed to create policy definition {}! ({})".format(
        name, policy.get("error")
    )
    return ret


def policy_definition_absent(name, connection_auth=None):
    """
    .. versionadded:: 2019.2.0

    Ensure a policy definition does not exist in the current subscription.

    :param name:
        Name of the policy definition.

    :param connection_auth:
        A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.
    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        ret[
            "comment"
        ] = "Connection information must be specified via connection_auth dictionary!"
        return ret

    policy = __salt__["azurearm_resource.policy_definition_get"](
        name, azurearm_log_level="info", **connection_auth
    )

    if "error" in policy:
        ret["result"] = True
        ret["comment"] = "Policy definition {} is already absent.".format(name)
        return ret

    elif __opts__["test"]:
        ret["comment"] = "Policy definition {} would be deleted.".format(name)
        ret["result"] = None
        ret["changes"] = {
            "old": policy,
            "new": {},
        }
        return ret

    deleted = __salt__["azurearm_resource.policy_definition_delete"](
        name, **connection_auth
    )

    if deleted:
        ret["result"] = True
        ret["comment"] = "Policy definition {} has been deleted.".format(name)
        ret["changes"] = {"old": policy, "new": {}}
        return ret

    ret["comment"] = "Failed to delete policy definition {}!".format(name)
    return ret


def policy_assignment_present(
    name,
    scope,
    definition_name,
    display_name=None,
    description=None,
    assignment_type=None,
    parameters=None,
    connection_auth=None,
    **kwargs
):
    """
    .. versionadded:: 2019.2.0

    Ensure a security policy assignment exists.

    :param name:
        Name of the policy assignment.

    :param scope:
        The scope of the policy assignment.

    :param definition_name:
        The name of the policy definition to assign.

    :param display_name:
        The display name of the policy assignment.

    :param description:
        The policy assignment description.

    :param assignment_type:
        The type of policy assignment.

    :param parameters:
        Required dictionary if a parameter is used in the policy rule.

    :param connection_auth:
        A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.

    Example usage:

    .. code-block:: yaml

        Ensure policy assignment exists:
            azurearm_resource.policy_assignment_present:
                - name: testassign
                - scope: /subscriptions/bc75htn-a0fhsi-349b-56gh-4fghti-f84852
                - definition_name: testpolicy
                - display_name: Test Assignment
                - description: Test assignment for testing assignments.
                - connection_auth: {{ profile }}

    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        ret[
            "comment"
        ] = "Connection information must be specified via connection_auth dictionary!"
        return ret

    policy = __salt__["azurearm_resource.policy_assignment_get"](
        name, scope, azurearm_log_level="info", **connection_auth
    )

    if "error" not in policy:
        if (
            assignment_type
            and assignment_type.lower() != policy.get("type", "").lower()
        ):
            ret["changes"]["type"] = {"old": policy.get("type"), "new": assignment_type}

        if scope.lower() != policy["scope"].lower():
            ret["changes"]["scope"] = {"old": policy["scope"], "new": scope}

        pa_name = policy["policy_definition_id"].split("/")[-1]
        if definition_name.lower() != pa_name.lower():
            ret["changes"]["definition_name"] = {"old": pa_name, "new": definition_name}

        if (display_name or "").lower() != policy.get("display_name", "").lower():
            ret["changes"]["display_name"] = {
                "old": policy.get("display_name"),
                "new": display_name,
            }

        if (description or "").lower() != policy.get("description", "").lower():
            ret["changes"]["description"] = {
                "old": policy.get("description"),
                "new": description,
            }

        param_changes = __utils__["dictdiffer.deep_diff"](
            policy.get("parameters", {}), parameters or {}
        )
        if param_changes:
            ret["changes"]["parameters"] = param_changes

        if not ret["changes"]:
            ret["result"] = True
            ret["comment"] = "Policy assignment {} is already present.".format(name)
            return ret

        if __opts__["test"]:
            ret["comment"] = "Policy assignment {} would be updated.".format(name)
            ret["result"] = None
            return ret

    else:
        ret["changes"] = {
            "old": {},
            "new": {
                "name": name,
                "scope": scope,
                "definition_name": definition_name,
                "type": assignment_type,
                "display_name": display_name,
                "description": description,
                "parameters": parameters,
            },
        }

    if __opts__["test"]:
        ret["comment"] = "Policy assignment {} would be created.".format(name)
        ret["result"] = None
        return ret

    if isinstance(parameters, dict):
        parameters = json.loads(json.dumps(parameters))

    policy_kwargs = kwargs.copy()
    policy_kwargs.update(connection_auth)
    policy = __salt__["azurearm_resource.policy_assignment_create"](
        name=name,
        scope=scope,
        definition_name=definition_name,
        type=assignment_type,
        display_name=display_name,
        description=description,
        parameters=parameters,
        **policy_kwargs
    )

    if "error" not in policy:
        ret["result"] = True
        ret["comment"] = "Policy assignment {} has been created.".format(name)
        return ret

    ret["comment"] = "Failed to create policy assignment {}! ({})".format(
        name, policy.get("error")
    )
    return ret


def policy_assignment_absent(name, scope, connection_auth=None):
    """
    .. versionadded:: 2019.2.0

    Ensure a policy assignment does not exist in the provided scope.

    :param name:
        Name of the policy assignment.

    :param scope:
        The scope of the policy assignment.

    connection_auth
        A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.
    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        ret[
            "comment"
        ] = "Connection information must be specified via connection_auth dictionary!"
        return ret

    policy = __salt__["azurearm_resource.policy_assignment_get"](
        name, scope, azurearm_log_level="info", **connection_auth
    )

    if "error" in policy:
        ret["result"] = True
        ret["comment"] = "Policy assignment {} is already absent.".format(name)
        return ret

    elif __opts__["test"]:
        ret["comment"] = "Policy assignment {} would be deleted.".format(name)
        ret["result"] = None
        ret["changes"] = {
            "old": policy,
            "new": {},
        }
        return ret

    deleted = __salt__["azurearm_resource.policy_assignment_delete"](
        name, scope, **connection_auth
    )

    if deleted:
        ret["result"] = True
        ret["comment"] = "Policy assignment {} has been deleted.".format(name)
        ret["changes"] = {"old": policy, "new": {}}
        return ret

    ret["comment"] = "Failed to delete policy assignment {}!".format(name)
    return ret



"""
Azure (ARM) DNS Execution Module

.. versionadded:: 3000

:maintainer: <devops@eitr.tech>
:maturity: new
:depends:
    * `azure <https://pypi.python.org/pypi/azure>`_ >= 2.0.0
    * `azure-common <https://pypi.python.org/pypi/azure-common>`_ >= 1.1.8
    * `azure-mgmt <https://pypi.python.org/pypi/azure-mgmt>`_ >= 1.0.0
    * `azure-mgmt-compute <https://pypi.python.org/pypi/azure-mgmt-compute>`_ >= 1.0.0
    * `azure-mgmt-dns <https://pypi.python.org/pypi/azure-mgmt-dns>`_ >= 2.0.0rc1
    * `azure-mgmt-network <https://pypi.python.org/pypi/azure-mgmt-network>`_ >= 1.7.1
    * `azure-mgmt-resource <https://pypi.python.org/pypi/azure-mgmt-resource>`_ >= 1.1.0
    * `azure-mgmt-storage <https://pypi.python.org/pypi/azure-mgmt-storage>`_ >= 1.0.0
    * `azure-mgmt-web <https://pypi.python.org/pypi/azure-mgmt-web>`_ >= 0.32.0
    * `azure-storage <https://pypi.python.org/pypi/azure-storage>`_ >= 0.34.3
    * `msrestazure <https://pypi.python.org/pypi/msrestazure>`_ >= 0.4.21

:platform: linux
:configuration:
    This module requires Azure Resource Manager credentials to be passed as keyword arguments
    to every function in order to work properly.

Required provider parameters:

    if using username and password:

        * ``subscription_id``
        * ``username``
        * ``password``

    if using a service principal:

        * ``subscription_id``
        * ``tenant``
        * ``client_id``
        * ``secret``

Optional provider parameters:

    **cloud_environment**: Used to point the cloud driver to different API endpoints, such as Azure GovCloud.

    Possible values:

        * ``AZURE_PUBLIC_CLOUD`` (default)
        * ``AZURE_CHINA_CLOUD``
        * ``AZURE_US_GOV_CLOUD``
        * ``AZURE_GERMAN_CLOUD``

"""

# Python libs

import logging

# Azure libs
HAS_LIBS = False
try:
    import azure.mgmt.dns.models  # pylint: disable=unused-import
    from msrest.exceptions import SerializationError
    from msrestazure.azure_exceptions import CloudError

    HAS_LIBS = True
except ImportError:
    pass

__virtualname__ = "azurearm_dns"

log = logging.getLogger(__name__)


def __virtual__():
    if not HAS_LIBS:
        return (
            False,
            "The following dependencies are required to use the AzureARM modules: "
            "Microsoft Azure SDK for Python >= 2.0rc6, "
            "MS REST Azure (msrestazure) >= 0.4",
        )

    return __virtualname__


def record_set_create_or_update(name, zone_name, resource_group, record_type, **kwargs):
    """
    .. versionadded:: 3000

    Creates or updates a record set within a DNS zone.

    :param name: The name of the record set, relative to the name of the zone.

    :param zone_name: The name of the DNS zone (without a terminating dot).

    :param resource_group: The name of the resource group.

    :param record_type:
        The type of DNS record in this record set. Record sets of type SOA can be
        updated but not created (they are created when the DNS zone is created).
        Possible values include: 'A', 'AAAA', 'CAA', 'CNAME', 'MX', 'NS', 'PTR', 'SOA', 'SRV', 'TXT'

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_dns.record_set_create_or_update myhost myzone testgroup A
            arecords='[{ipv4_address: 10.0.0.1}]' ttl=300

    """
    dnsconn = __utils__["azurearm.get_client"]("dns", **kwargs)

    try:
        record_set_model = __utils__["azurearm.create_object_model"](
            "dns", "RecordSet", **kwargs
        )
    except TypeError as exc:
        result = {"error": "The object model could not be built. ({})".format(str(exc))}
        return result

    try:
        record_set = dnsconn.record_sets.create_or_update(
            relative_record_set_name=name,
            zone_name=zone_name,
            resource_group_name=resource_group,
            record_type=record_type,
            parameters=record_set_model,
            if_match=kwargs.get("if_match"),
            if_none_match=kwargs.get("if_none_match"),
        )
        result = record_set.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("dns", str(exc), **kwargs)
        result = {"error": str(exc)}
    except SerializationError as exc:
        result = {
            "error": "The object model could not be parsed. ({})".format(str(exc))
        }

    return result


def record_set_delete(name, zone_name, resource_group, record_type, **kwargs):
    """
    .. versionadded:: 3000

    Deletes a record set from a DNS zone. This operation cannot be undone.

    :param name: The name of the record set, relative to the name of the zone.

    :param zone_name: The name of the DNS zone (without a terminating dot).

    :param resource_group: The name of the resource group.

    :param record_type:
        The type of DNS record in this record set. Record sets of type SOA cannot be
        deleted (they are deleted when the DNS zone is deleted).
        Possible values include: 'A', 'AAAA', 'CAA', 'CNAME', 'MX', 'NS', 'PTR', 'SOA', 'SRV', 'TXT'

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_dns.record_set_delete myhost myzone testgroup A

    """
    result = False
    dnsconn = __utils__["azurearm.get_client"]("dns", **kwargs)
    try:
        record_set = dnsconn.record_sets.delete(
            relative_record_set_name=name,
            zone_name=zone_name,
            resource_group_name=resource_group,
            record_type=record_type,
            if_match=kwargs.get("if_match"),
        )
        result = True
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("dns", str(exc), **kwargs)

    return result


def record_set_get(name, zone_name, resource_group, record_type, **kwargs):
    """
    .. versionadded:: 3000

    Get a dictionary representing a record set's properties.

    :param name: The name of the record set, relative to the name of the zone.

    :param zone_name: The name of the DNS zone (without a terminating dot).

    :param resource_group: The name of the resource group.

    :param record_type:
        The type of DNS record in this record set.
        Possible values include: 'A', 'AAAA', 'CAA', 'CNAME', 'MX', 'NS', 'PTR', 'SOA', 'SRV', 'TXT'

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_dns.record_set_get '@' myzone testgroup SOA

    """
    dnsconn = __utils__["azurearm.get_client"]("dns", **kwargs)
    try:
        record_set = dnsconn.record_sets.get(
            relative_record_set_name=name,
            zone_name=zone_name,
            resource_group_name=resource_group,
            record_type=record_type,
        )
        result = record_set.as_dict()

    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("dns", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


def record_sets_list_by_type(
    zone_name, resource_group, record_type, top=None, recordsetnamesuffix=None, **kwargs
):
    """
    .. versionadded:: 3000

    Lists the record sets of a specified type in a DNS zone.

    :param zone_name: The name of the DNS zone (without a terminating dot).

    :param resource_group: The name of the resource group.

    :param record_type:
        The type of record sets to enumerate.
        Possible values include: 'A', 'AAAA', 'CAA', 'CNAME', 'MX', 'NS', 'PTR', 'SOA', 'SRV', 'TXT'

    :param top:
        The maximum number of record sets to return. If not specified,
        returns up to 100 record sets.

    :param recordsetnamesuffix:
        The suffix label of the record set name that has
        to be used to filter the record set enumerations.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_dns.record_sets_list_by_type myzone testgroup SOA

    """
    result = {}
    dnsconn = __utils__["azurearm.get_client"]("dns", **kwargs)
    try:
        record_sets = __utils__["azurearm.paged_object_to_list"](
            dnsconn.record_sets.list_by_type(
                zone_name=zone_name,
                resource_group_name=resource_group,
                record_type=record_type,
                top=top,
                recordsetnamesuffix=recordsetnamesuffix,
            )
        )

        for record_set in record_sets:
            result[record_set["name"]] = record_set
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("dns", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


def record_sets_list_by_dns_zone(
    zone_name, resource_group, top=None, recordsetnamesuffix=None, **kwargs
):
    """
    .. versionadded:: 3000

    Lists all record sets in a DNS zone.

    :param zone_name: The name of the DNS zone (without a terminating dot).

    :param resource_group: The name of the resource group.

    :param top:
        The maximum number of record sets to return. If not specified,
        returns up to 100 record sets.

    :param recordsetnamesuffix:
        The suffix label of the record set name that has
        to be used to filter the record set enumerations.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_dns.record_sets_list_by_dns_zone myzone testgroup

    """
    result = {}
    dnsconn = __utils__["azurearm.get_client"]("dns", **kwargs)
    try:
        record_sets = __utils__["azurearm.paged_object_to_list"](
            dnsconn.record_sets.list_by_dns_zone(
                zone_name=zone_name,
                resource_group_name=resource_group,
                top=top,
                recordsetnamesuffix=recordsetnamesuffix,
            )
        )

        for record_set in record_sets:
            result[record_set["name"]] = record_set
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("dns", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


def zone_create_or_update(name, resource_group, **kwargs):
    """
    .. versionadded:: 3000

    Creates or updates a DNS zone. Does not modify DNS records within the zone.

    :param name: The name of the DNS zone to create (without a terminating dot).

    :param resource_group: The name of the resource group.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_dns.zone_create_or_update myzone testgroup

    """
    # DNS zones are global objects
    kwargs["location"] = "global"

    dnsconn = __utils__["azurearm.get_client"]("dns", **kwargs)

    # Convert list of ID strings to list of dictionaries with id key.
    if isinstance(kwargs.get("registration_virtual_networks"), list):
        kwargs["registration_virtual_networks"] = [
            {"id": vnet} for vnet in kwargs["registration_virtual_networks"]
        ]

    if isinstance(kwargs.get("resolution_virtual_networks"), list):
        kwargs["resolution_virtual_networks"] = [
            {"id": vnet} for vnet in kwargs["resolution_virtual_networks"]
        ]

    try:
        zone_model = __utils__["azurearm.create_object_model"]("dns", "Zone", **kwargs)
    except TypeError as exc:
        result = {"error": "The object model could not be built. ({})".format(str(exc))}
        return result

    try:
        zone = dnsconn.zones.create_or_update(
            zone_name=name,
            resource_group_name=resource_group,
            parameters=zone_model,
            if_match=kwargs.get("if_match"),
            if_none_match=kwargs.get("if_none_match"),
        )
        result = zone.as_dict()
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("dns", str(exc), **kwargs)
        result = {"error": str(exc)}
    except SerializationError as exc:
        result = {
            "error": "The object model could not be parsed. ({})".format(str(exc))
        }

    return result


def zone_delete(name, resource_group, **kwargs):
    """
    .. versionadded:: 3000

    Delete a DNS zone within a resource group.

    :param name: The name of the DNS zone to delete.

    :param resource_group: The name of the resource group.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_dns.zone_delete myzone testgroup

    """
    result = False
    dnsconn = __utils__["azurearm.get_client"]("dns", **kwargs)
    try:
        zone = dnsconn.zones.delete(
            zone_name=name,
            resource_group_name=resource_group,
            if_match=kwargs.get("if_match"),
        )
        zone.wait()
        result = True
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("dns", str(exc), **kwargs)

    return result


def zone_get(name, resource_group, **kwargs):
    """
    .. versionadded:: 3000

    Get a dictionary representing a DNS zone's properties, but not the
    record sets within the zone.

    :param name: The DNS zone to get.

    :param resource_group: The name of the resource group.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_dns.zone_get myzone testgroup

    """
    dnsconn = __utils__["azurearm.get_client"]("dns", **kwargs)
    try:
        zone = dnsconn.zones.get(zone_name=name, resource_group_name=resource_group)
        result = zone.as_dict()

    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("dns", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


def zones_list_by_resource_group(resource_group, top=None, **kwargs):
    """
    .. versionadded:: 3000

    Lists the DNS zones in a resource group.

    :param resource_group: The name of the resource group.

    :param top:
        The maximum number of DNS zones to return. If not specified,
        returns up to 100 zones.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_dns.zones_list_by_resource_group testgroup

    """
    result = {}
    dnsconn = __utils__["azurearm.get_client"]("dns", **kwargs)
    try:
        zones = __utils__["azurearm.paged_object_to_list"](
            dnsconn.zones.list_by_resource_group(
                resource_group_name=resource_group, top=top
            )
        )

        for zone in zones:
            result[zone["name"]] = zone
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("dns", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result


def zones_list(top=None, **kwargs):
    """
    .. versionadded:: 3000

    Lists the DNS zones in all resource groups in a subscription.

    :param top:
        The maximum number of DNS zones to return. If not specified,
        eturns up to 100 zones.

    CLI Example:

    .. code-block:: bash

        salt-call azurearm_dns.zones_list

    """
    result = {}
    dnsconn = __utils__["azurearm.get_client"]("dns", **kwargs)
    try:
        zones = __utils__["azurearm.paged_object_to_list"](dnsconn.zones.list(top=top))

        for zone in zones:
            result[zone["name"]] = zone
    except CloudError as exc:
        __utils__["azurearm.log_cloud_error"]("dns", str(exc), **kwargs)
        result = {"error": str(exc)}

    return result



"""
Dynamic DNS Runner
==================

.. versionadded:: 2015.8.0

Runner to interact with DNS server and create/delete/update DNS records

:codeauthor: Nitin Madhok <nmadhok@g.clemson.edu>

"""

import logging
import os

import salt.utils.files
import salt.utils.json

HAS_LIBS = False
try:
    import dns.query
    import dns.update
    import dns.tsigkeyring

    HAS_LIBS = True
except ImportError:
    HAS_LIBS = False


log = logging.getLogger(__name__)


def __virtual__():
    """
    Check if required libs (python-dns) is installed and load runner
    only if they are present
    """
    if not HAS_LIBS:
        return False

    return True


def _get_keyring(keyfile):
    keyring = None
    if keyfile and os.path.isfile(os.path.expanduser(keyfile)):
        with salt.utils.files.fopen(keyfile) as _f:
            keyring = dns.tsigkeyring.from_text(salt.utils.json.load(_f))

    return keyring


def create(
    zone,
    name,
    ttl,
    rdtype,
    data,
    keyname,
    keyfile,
    nameserver,
    timeout,
    port=53,
    keyalgorithm="hmac-md5",
):
    """
    Create a DNS record. The nameserver must be an IP address and the master running
    this runner must have create privileges on that server.

    CLI Example:

    .. code-block:: bash

        salt-run ddns.create domain.com my-test-vm 3600 A 10.20.30.40 my-tsig-key /etc/salt/tsig.keyring 10.0.0.1 5
    """
    if zone in name:
        name = name.replace(zone, "").rstrip(".")
    fqdn = "{}.{}".format(name, zone)
    request = dns.message.make_query(fqdn, rdtype)
    answer = dns.query.udp(request, nameserver, timeout, port)

    rdata_value = dns.rdatatype.from_text(rdtype)
    rdata = dns.rdata.from_text(dns.rdataclass.IN, rdata_value, data)

    for rrset in answer.answer:
        if rdata in rrset.items:
            return {
                fqdn: "Record of type '{}' already exists with ttl of {}".format(
                    rdtype, rrset.ttl
                )
            }

    keyring = _get_keyring(keyfile)

    dns_update = dns.update.Update(
        zone, keyring=keyring, keyname=keyname, keyalgorithm=keyalgorithm
    )
    dns_update.add(name, ttl, rdata)

    answer = dns.query.udp(dns_update, nameserver, timeout, port)
    if answer.rcode() > 0:
        return {fqdn: "Failed to create record of type '{}'".format(rdtype)}

    return {fqdn: "Created record of type '{}': {} -> {}".format(rdtype, fqdn, data)}


def update(
    zone,
    name,
    ttl,
    rdtype,
    data,
    keyname,
    keyfile,
    nameserver,
    timeout,
    replace=False,
    port=53,
    keyalgorithm="hmac-md5",
):
    """
    Replace, or update a DNS record. The nameserver must be an IP address and the master running
    this runner must have update privileges on that server.

    .. note::

        If ``replace`` is set to True, all records for this name and type will first be deleted and
        then recreated. Default is ``replace=False``.

    CLI Example:

    .. code-block:: bash

        salt-run ddns.update domain.com my-test-vm 3600 A 10.20.30.40 my-tsig-key /etc/salt/tsig.keyring 10.0.0.1 5
    """
    if zone in name:
        name = name.replace(zone, "").rstrip(".")
    fqdn = "{}.{}".format(name, zone)
    request = dns.message.make_query(fqdn, rdtype)
    answer = dns.query.udp(request, nameserver, timeout, port)
    if not answer.answer:
        return {fqdn: "No matching DNS record(s) found"}

    rdata_value = dns.rdatatype.from_text(rdtype)
    rdata = dns.rdata.from_text(dns.rdataclass.IN, rdata_value, data)

    for rrset in answer.answer:
        if rdata in rrset.items:
            rr = rrset.items
            if ttl == rrset.ttl:
                if replace and (len(answer.answer) > 1 or len(rrset.items) > 1):
                    break
                return {
                    fqdn: "Record of type '{}' already present with ttl of {}".format(
                        rdtype, ttl
                    )
                }
            break

    keyring = _get_keyring(keyfile)

    dns_update = dns.update.Update(
        zone, keyring=keyring, keyname=keyname, keyalgorithm=keyalgorithm
    )
    dns_update.replace(name, ttl, rdata)

    answer = dns.query.udp(dns_update, nameserver, timeout, port)
    if answer.rcode() > 0:
        return {fqdn: "Failed to update record of type '{}'".format(rdtype)}

    return {fqdn: "Updated record of type '{}'".format(rdtype)}


def delete(
    zone,
    name,
    keyname,
    keyfile,
    nameserver,
    timeout,
    rdtype=None,
    data=None,
    port=53,
    keyalgorithm="hmac-md5",
):
    """
    Delete a DNS record.

    CLI Example:

    .. code-block:: bash

        salt-run ddns.delete domain.com my-test-vm my-tsig-key /etc/salt/tsig.keyring 10.0.0.1 5 A
    """
    if zone in name:
        name = name.replace(zone, "").rstrip(".")
    fqdn = "{}.{}".format(name, zone)
    request = dns.message.make_query(fqdn, (rdtype or "ANY"))

    answer = dns.query.udp(request, nameserver, timeout, port)
    if not answer.answer:
        return {fqdn: "No matching DNS record(s) found"}

    keyring = _get_keyring(keyfile)

    dns_update = dns.update.Update(
        zone, keyring=keyring, keyname=keyname, keyalgorithm=keyalgorithm
    )

    if rdtype:
        rdata_value = dns.rdatatype.from_text(rdtype)
        if data:
            rdata = dns.rdata.from_text(dns.rdataclass.IN, rdata_value, data)
            dns_update.delete(name, rdata)
        else:
            dns_update.delete(name, rdata_value)
    else:
        dns_update.delete(name)

    answer = dns.query.udp(dns_update, nameserver, timeout, port)
    if answer.rcode() > 0:
        return {fqdn: "Failed to delete DNS record(s)"}

    return {fqdn: "Deleted DNS record(s)"}


def add_host(
    zone,
    name,
    ttl,
    ip,
    keyname,
    keyfile,
    nameserver,
    timeout,
    port=53,
    keyalgorithm="hmac-md5",
):
    """
    Create both A and PTR (reverse) records for a host.

    CLI Example:

    .. code-block:: bash

        salt-run ddns.add_host domain.com my-test-vm 3600 10.20.30.40 my-tsig-key /etc/salt/tsig.keyring 10.0.0.1 5
    """
    res = []
    if zone in name:
        name = name.replace(zone, "").rstrip(".")
    fqdn = "{}.{}".format(name, zone)

    ret = create(
        zone,
        name,
        ttl,
        "A",
        ip,
        keyname,
        keyfile,
        nameserver,
        timeout,
        port,
        keyalgorithm,
    )
    res.append(ret[fqdn])

    parts = ip.split(".")[::-1]
    i = len(parts)
    popped = []

    # Iterate over possible reverse zones
    while i > 1:
        p = parts.pop(0)
        i -= 1
        popped.append(p)

        zone = "{}.{}".format(".".join(parts), "in-addr.arpa.")
        name = ".".join(popped)
        rev_fqdn = "{}.{}".format(name, zone)
        ret = create(
            zone,
            name,
            ttl,
            "PTR",
            "{}.".format(fqdn),
            keyname,
            keyfile,
            nameserver,
            timeout,
            port,
            keyalgorithm,
        )

        if "Created" in ret[rev_fqdn]:
            res.append(ret[rev_fqdn])
            return {fqdn: res}

    res.append(ret[rev_fqdn])

    return {fqdn: res}


def delete_host(
    zone, name, keyname, keyfile, nameserver, timeout, port=53, keyalgorithm="hmac-md5"
):
    """
    Delete both forward (A) and reverse (PTR) records for a host only if the
    forward (A) record exists.

    CLI Example:

    .. code-block:: bash

        salt-run ddns.delete_host domain.com my-test-vm my-tsig-key /etc/salt/tsig.keyring 10.0.0.1 5
    """
    res = []
    if zone in name:
        name = name.replace(zone, "").rstrip(".")
    fqdn = "{}.{}".format(name, zone)
    request = dns.message.make_query(fqdn, "A")
    answer = dns.query.udp(request, nameserver, timeout, port)

    try:
        ips = [i.address for i in answer.answer[0].items]
    except IndexError:
        ips = []

    ret = delete(
        zone,
        name,
        keyname,
        keyfile,
        nameserver,
        timeout,
        port=port,
        keyalgorithm=keyalgorithm,
    )
    res.append("{} of type 'A'".format(ret[fqdn]))

    for ip in ips:
        parts = ip.split(".")[::-1]
        i = len(parts)
        popped = []

        # Iterate over possible reverse zones
        while i > 1:
            p = parts.pop(0)
            i -= 1
            popped.append(p)
            zone = "{}.{}".format(".".join(parts), "in-addr.arpa.")
            name = ".".join(popped)
            rev_fqdn = "{}.{}".format(name, zone)
            ret = delete(
                zone,
                name,
                keyname,
                keyfile,
                nameserver,
                timeout,
                "PTR",
                "{}.".format(fqdn),
                port,
                keyalgorithm,
            )

            if "Deleted" in ret[rev_fqdn]:
                res.append("{} of type 'PTR'".format(ret[rev_fqdn]))
                return {fqdn: res}

        res.append(ret[rev_fqdn])

    return {fqdn: res}



"""
Module to provide Citrix Netscaler compatibility to Salt (compatible with netscaler 9.2+)

.. versionadded:: 2015.2.0

:depends:

- nsnitro Python module

.. note::
    You can install nsnitro using:

    .. code-block:: bash

        pip install nsnitro

:configuration: This module accepts connection configuration details either as
    parameters, or as configuration settings in /etc/salt/minion on the relevant
    minions

    .. code-block:: yaml

        netscaler.host: 1.2.3.4
        netscaler.user: user
        netscaler.pass: password

    This data can also be passed into pillar. Options passed into opts will
    overwrite options passed into pillar.

CLI Examples:

    Calls relying on configuration passed using /etc/salt/minion, grains, or pillars:

    .. code-block:: bash

        salt-call netscaler.server_exists server_name

    Calls passing configuration as opts

    .. code-block:: bash

        salt-call netscaler.server_exists server_name netscaler_host=1.2.3.4 netscaler_user=username netscaler_pass=password
        salt-call netscaler.server_exists server_name netscaler_host=1.2.3.5 netscaler_user=username2 netscaler_pass=password2
        salt-call netscaler.server_enable server_name2 netscaler_host=1.2.3.5
        salt-call netscaler.server_up server_name3 netscaler_host=1.2.3.6 netscaler_useSSL=False

"""

import logging

import salt.utils.platform

try:
    from nsnitro.nsnitro import NSNitro
    from nsnitro.nsexceptions import NSNitroError
    from nsnitro.nsresources.nsserver import NSServer
    from nsnitro.nsresources.nsservice import NSService
    from nsnitro.nsresources.nsservicegroup import NSServiceGroup
    from nsnitro.nsresources.nsservicegroupserverbinding import (
        NSServiceGroupServerBinding,
    )
    from nsnitro.nsresources.nslbvserver import NSLBVServer
    from nsnitro.nsresources.nslbvserverservicegroupbinding import (
        NSLBVServerServiceGroupBinding,
    )
    from nsnitro.nsresources.nssslvserversslcertkeybinding import (
        NSSSLVServerSSLCertKeyBinding,
    )

    HAS_NSNITRO = True
except ImportError:
    HAS_NSNITRO = False

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load this module if the nsnitro library is installed
    """
    if salt.utils.platform.is_windows():
        return (
            False,
            "The netscaler execution module failed to load: not available on Windows.",
        )
    if HAS_NSNITRO:
        return "netscaler"
    return (
        False,
        "The netscaler execution module failed to load: the nsnitro python "
        "library is not available.",
    )


def _connect(**kwargs):
    """
    Initialise netscaler connection
    """
    connargs = dict()

    # Shamelessy ripped from the mysql module
    def __connarg(name, key=None, default=None):
        """
        Add key to connargs, only if name exists in our kwargs or as
        netscaler.<name> in __opts__ or __pillar__ Evaluate in said order - kwargs,
        opts then pillar. To avoid collision with other functions, kwargs-based
        connection arguments are prefixed with 'netscaler_' (i.e.
        'netscaler_host', 'netscaler_user', etc.).
        """
        if key is None:
            key = name
        if name in kwargs:
            connargs[key] = kwargs[name]
        else:
            prefix = "netscaler_"
            if name.startswith(prefix):
                try:
                    name = name[len(prefix) :]
                except IndexError:
                    return
            val = __salt__["config.option"]("netscaler.{}".format(name), None)
            if val is not None:
                connargs[key] = val
            elif default is not None:
                connargs[key] = default

    __connarg("netscaler_host", "host")
    __connarg("netscaler_user", "user")
    __connarg("netscaler_pass", "pass")
    __connarg("netscaler_useSSL", "useSSL", True)

    nitro = NSNitro(
        connargs["host"], connargs["user"], connargs["pass"], connargs["useSSL"]
    )
    try:
        nitro.login()
    except NSNitroError as error:
        log.debug("netscaler module error - NSNitro.login() failed: %s", error)
        return None
    return nitro


def _disconnect(nitro):
    try:
        nitro.logout()
    except NSNitroError as error:
        log.debug("netscaler module error - NSNitro.logout() failed: %s", error)
        return None
    return nitro


def _servicegroup_get(sg_name, **connection_args):
    """
    Return a service group ressource or None
    """
    nitro = _connect(**connection_args)
    if nitro is None:
        return None
    sg = NSServiceGroup()
    sg.set_servicegroupname(sg_name)
    try:
        sg = NSServiceGroup.get(nitro, sg)
    except NSNitroError as error:
        log.debug("netscaler module error - NSServiceGroup.get() failed: %s", error)
        sg = None
    _disconnect(nitro)
    return sg


def _servicegroup_get_servers(sg_name, **connection_args):
    """
    Returns a list of members of a servicegroup or None
    """
    nitro = _connect(**connection_args)
    if nitro is None:
        return None
    sg = NSServiceGroup()
    sg.set_servicegroupname(sg_name)
    try:
        sg = NSServiceGroup.get_servers(nitro, sg)
    except NSNitroError as error:
        log.debug(
            "netscaler module error - NSServiceGroup.get_servers failed(): %s", error
        )
        sg = None
    _disconnect(nitro)
    return sg


def _servicegroup_get_server(sg_name, s_name, s_port=None, **connection_args):
    """
    Returns a member of a service group or None
    """
    ret = None
    servers = _servicegroup_get_servers(sg_name, **connection_args)
    if servers is None:
        return None
    for server in servers:
        if server.get_servername() == s_name:
            if s_port is not None and s_port != server.get_port():
                ret = None
            ret = server
    return ret


def servicegroup_exists(sg_name, sg_type=None, **connection_args):
    """
    Checks if a service group exists

    CLI Example:

    .. code-block:: bash

        salt '*' netscaler.servicegroup_exists 'serviceGroupName'
    """
    sg = _servicegroup_get(sg_name, **connection_args)
    if sg is None:
        return False
    if sg_type is not None and sg_type.upper() != sg.get_servicetype():
        return False
    return True


def servicegroup_add(sg_name, sg_type="HTTP", **connection_args):
    """
    Add a new service group
    If no service type is specified, HTTP will be used.
    Most common service types: HTTP, SSL, and SSL_BRIDGE

    CLI Example:

    .. code-block:: bash

        salt '*' netscaler.servicegroup_add 'serviceGroupName'
        salt '*' netscaler.servicegroup_add 'serviceGroupName' 'serviceGroupType'
    """
    ret = True
    if servicegroup_exists(sg_name):
        return False
    nitro = _connect(**connection_args)
    if nitro is None:
        return False
    sg = NSServiceGroup()
    sg.set_servicegroupname(sg_name)
    sg.set_servicetype(sg_type.upper())
    try:
        NSServiceGroup.add(nitro, sg)
    except NSNitroError as error:
        log.debug("netscaler module error - NSServiceGroup.add() failed: %s", error)
        ret = False
    _disconnect(nitro)
    return ret


def servicegroup_delete(sg_name, **connection_args):
    """
    Delete a new service group

    CLI Example:

    .. code-block:: bash

        salt '*' netscaler.servicegroup_delete 'serviceGroupName'
    """
    ret = True
    sg = _servicegroup_get(sg_name, **connection_args)
    if sg is None:
        return False
    nitro = _connect(**connection_args)
    if nitro is None:
        return False
    try:
        NSServiceGroup.delete(nitro, sg)
    except NSNitroError as error:
        log.debug("netscaler module error - NSServiceGroup.delete() failed: %s", error)
        ret = False
    _disconnect(nitro)
    return ret


def servicegroup_server_exists(sg_name, s_name, s_port=None, **connection_args):
    """
    Check if a server:port combination is a member of a servicegroup

    CLI Example:

    .. code-block:: bash

        salt '*' netscaler.servicegroup_server_exists 'serviceGroupName' 'serverName' 'serverPort'
    """
    return (
        _servicegroup_get_server(sg_name, s_name, s_port, **connection_args) is not None
    )


def servicegroup_server_up(sg_name, s_name, s_port, **connection_args):
    """
    Check if a server:port combination is in state UP in a servicegroup

    CLI Example:

    .. code-block:: bash

        salt '*' netscaler.servicegroup_server_up 'serviceGroupName' 'serverName' 'serverPort'
    """
    server = _servicegroup_get_server(sg_name, s_name, s_port, **connection_args)
    return server is not None and server.get_svrstate() == "UP"


def servicegroup_server_enable(sg_name, s_name, s_port, **connection_args):
    """
    Enable a server:port member of a servicegroup

    CLI Example:

    .. code-block:: bash

        salt '*' netscaler.servicegroup_server_enable 'serviceGroupName' 'serverName' 'serverPort'
    """
    ret = True
    server = _servicegroup_get_server(sg_name, s_name, s_port, **connection_args)
    if server is None:
        return False
    nitro = _connect(**connection_args)
    if nitro is None:
        return False
    try:
        NSServiceGroup.enable_server(nitro, server)
    except NSNitroError as error:
        log.debug(
            "netscaler module error - NSServiceGroup.enable_server() failed: %s", error
        )
        ret = False
    _disconnect(nitro)
    return ret


def servicegroup_server_disable(sg_name, s_name, s_port, **connection_args):
    """
    Disable a server:port member of a servicegroup

    CLI Example:

    .. code-block:: bash

        salt '*' netscaler.servicegroup_server_disable 'serviceGroupName' 'serverName' 'serverPort'
    """
    ret = True
    server = _servicegroup_get_server(sg_name, s_name, s_port, **connection_args)
    if server is None:
        return False
    nitro = _connect(**connection_args)
    if nitro is None:
        return False
    try:
        NSServiceGroup.disable_server(nitro, server)
    except NSNitroError as error:
        log.debug(
            "netscaler module error - NSServiceGroup.disable_server() failed: %s", error
        )
        ret = False
    _disconnect(nitro)
    return ret


def servicegroup_server_add(sg_name, s_name, s_port, **connection_args):
    """
    Add a server:port member to a servicegroup

    CLI Example:

    .. code-block:: bash

        salt '*' netscaler.servicegroup_server_add 'serviceGroupName' 'serverName' 'serverPort'
    """
    # Nitro will throw an error if the server is already present
    ret = True
    server = _servicegroup_get_server(sg_name, s_name, s_port, **connection_args)
    if server is not None:
        return False
    nitro = _connect(**connection_args)
    if nitro is None:
        return False
    sgsb = NSServiceGroupServerBinding()
    sgsb.set_servicegroupname(sg_name)
    sgsb.set_servername(s_name)
    sgsb.set_port(s_port)
    try:
        NSServiceGroupServerBinding.add(nitro, sgsb)
    except NSNitroError as error:
        log.debug(
            "netscaler module error - NSServiceGroupServerBinding() failed: %s", error
        )
        ret = False
    _disconnect(nitro)
    return ret


def servicegroup_server_delete(sg_name, s_name, s_port, **connection_args):
    """
    Remove a server:port member from a servicegroup

    CLI Example:

    .. code-block:: bash

        salt '*' netscaler.servicegroup_server_delete 'serviceGroupName' 'serverName' 'serverPort'
    """
    # Nitro will throw an error if the server is already not present
    ret = True
    server = _servicegroup_get_server(sg_name, s_name, s_port, **connection_args)
    if server is None:
        return False
    nitro = _connect(**connection_args)
    if nitro is None:
        return False
    sgsb = NSServiceGroupServerBinding()
    sgsb.set_servicegroupname(sg_name)
    sgsb.set_servername(s_name)
    sgsb.set_port(s_port)
    try:
        NSServiceGroupServerBinding.delete(nitro, sgsb)
    except NSNitroError as error:
        log.debug(
            "netscaler module error - NSServiceGroupServerBinding() failed: %s", error
        )
        ret = False
    _disconnect(nitro)
    return ret


def _service_get(s_name, **connection_args):
    """
    Returns a service ressource or None
    """
    nitro = _connect(**connection_args)
    if nitro is None:
        return None
    service = NSService()
    service.set_name(s_name)
    try:
        service = NSService.get(nitro, service)
    except NSNitroError as error:
        log.debug("netscaler module error - NSService.get() failed: %s", error)
        service = None
    _disconnect(nitro)
    return service


def service_exists(s_name, **connection_args):
    """
    Checks if a service exists

    CLI Example:

    .. code-block:: bash

        salt '*' netscaler.service_exists 'serviceName'
    """
    return _service_get(s_name, **connection_args) is not None


def service_up(s_name, **connection_args):
    """
    Checks if a service is UP

    CLI Example:

    .. code-block:: bash

        salt '*' netscaler.service_up 'serviceName'
    """
    service = _service_get(s_name, **connection_args)
    return service is not None and service.get_svrstate() == "UP"


def service_enable(s_name, **connection_args):
    """
    Enable a service

    CLI Example:

    .. code-block:: bash

        salt '*' netscaler.service_enable 'serviceName'
    """
    ret = True
    service = _service_get(s_name, **connection_args)
    if service is None:
        return False
    nitro = _connect(**connection_args)
    if nitro is None:
        return False
    try:
        NSService.enable(nitro, service)
    except NSNitroError as error:
        log.debug("netscaler module error - NSService.enable() failed: %s", error)
        ret = False
    _disconnect(nitro)
    return ret


def service_disable(s_name, s_delay=None, **connection_args):
    """
    Disable a service

    CLI Example:

    .. code-block:: bash

        salt '*' netscaler.service_disable 'serviceName'
        salt '*' netscaler.service_disable 'serviceName' 'delayInSeconds'
    """
    ret = True
    service = _service_get(s_name, **connection_args)
    if service is None:
        return False
    if s_delay is not None:
        service.set_delay(s_delay)
    nitro = _connect(**connection_args)
    if nitro is None:
        return False
    try:
        NSService.disable(nitro, service)
    except NSNitroError as error:
        log.debug("netscaler module error - NSService.enable() failed: %s", error)
        ret = False
    _disconnect(nitro)
    return ret


def _server_get(s_name, **connection_args):
    nitro = _connect(**connection_args)
    if nitro is None:
        return None
    server = NSServer()
    server.set_name(s_name)
    try:
        server = NSServer.get(nitro, server)
    except NSNitroError as error:
        log.debug("netscaler module error - NSServer.get() failed: %s", error)
        server = None
    _disconnect(nitro)
    return server


def server_exists(s_name, ip=None, s_state=None, **connection_args):
    """
    Checks if a server exists

    CLI Example:

    .. code-block:: bash

        salt '*' netscaler.server_exists 'serverName'
    """
    server = _server_get(s_name, **connection_args)
    if server is None:
        return False
    if ip is not None and ip != server.get_ipaddress():
        return False
    if s_state is not None and s_state.upper() != server.get_state():
        return False
    return True


def server_add(s_name, s_ip, s_state=None, **connection_args):
    """
    Add a server
    Note: The default server state is ENABLED

    CLI Example:

    .. code-block:: bash

        salt '*' netscaler.server_add 'serverName' 'serverIpAddress'
        salt '*' netscaler.server_add 'serverName' 'serverIpAddress' 'serverState'
    """
    ret = True
    if server_exists(s_name, **connection_args):
        return False
    nitro = _connect(**connection_args)
    if nitro is None:
        return False
    server = NSServer()
    server.set_name(s_name)
    server.set_ipaddress(s_ip)
    if s_state is not None:
        server.set_state(s_state)
    try:
        NSServer.add(nitro, server)
    except NSNitroError as error:
        log.debug("netscaler module error - NSServer.add() failed: %s", error)
        ret = False
    _disconnect(nitro)
    return ret


def server_delete(s_name, **connection_args):
    """
    Delete a server

    CLI Example:

    .. code-block:: bash

        salt '*' netscaler.server_delete 'serverName'
    """
    ret = True
    server = _server_get(s_name, **connection_args)
    if server is None:
        return False
    nitro = _connect(**connection_args)
    if nitro is None:
        return False
    try:
        NSServer.delete(nitro, server)
    except NSNitroError as error:
        log.debug("netscaler module error - NSServer.delete() failed: %s", error)
        ret = False
    _disconnect(nitro)
    return ret


def server_update(s_name, s_ip, **connection_args):
    """
    Update a server's attributes

    CLI Example:

    .. code-block:: bash

      salt '*' netscaler.server_update 'serverName' 'serverIP'
    """
    altered = False
    cur_server = _server_get(s_name, **connection_args)
    if cur_server is None:
        return False
    alt_server = NSServer()
    alt_server.set_name(s_name)
    if cur_server.get_ipaddress() != s_ip:
        alt_server.set_ipaddress(s_ip)
        altered = True
    # Nothing to update, the server is already idem
    if altered is False:
        return False
    # Perform the update
    nitro = _connect(**connection_args)
    if nitro is None:
        return False
    ret = True
    try:
        NSServer.update(nitro, alt_server)
    except NSNitroError as error:
        log.debug("netscaler module error - NSServer.update() failed: %s", error)
        ret = False
    _disconnect(nitro)
    return ret


def server_enabled(s_name, **connection_args):
    """
    Check if a server is enabled globally

    CLI Example:

    .. code-block:: bash

        salt '*' netscaler.server_enabled 'serverName'
    """
    server = _server_get(s_name, **connection_args)
    return server is not None and server.get_state() == "ENABLED"


def server_enable(s_name, **connection_args):
    """
    Enables a server globally

    CLI Example:

    .. code-block:: bash

        salt '*' netscaler.server_enable 'serverName'
    """
    ret = True
    server = _server_get(s_name, **connection_args)
    if server is None:
        return False
    if server.get_state() == "ENABLED":
        return True
    nitro = _connect(**connection_args)
    if nitro is None:
        return False
    try:
        NSServer.enable(nitro, server)
    except NSNitroError as error:
        log.debug("netscaler module error - NSServer.enable() failed: %s", error)
        ret = False
    _disconnect(nitro)
    return ret


def server_disable(s_name, **connection_args):
    """
    Disable a server globally

    CLI Example:

    .. code-block:: bash

        salt '*' netscaler.server_disable 'serverName'
    """
    ret = True
    server = _server_get(s_name, **connection_args)
    if server is None:
        return False
    if server.get_state() == "DISABLED":
        return True
    nitro = _connect(**connection_args)
    if nitro is None:
        return False
    try:
        NSServer.disable(nitro, server)
    except NSNitroError as error:
        log.debug("netscaler module error - NSServer.disable() failed: %s", error)
        ret = False
    _disconnect(nitro)
    return ret


def _vserver_get(v_name, **connection_args):
    nitro = _connect(**connection_args)
    vserver = NSLBVServer()
    vserver.set_name(v_name)
    if nitro is None:
        return None
    try:
        vserver = NSLBVServer.get(nitro, vserver)
    except NSNitroError as error:
        log.debug("netscaler module error - NSLBVServer.get() failed: %s", error)
        vserver = None
    _disconnect(nitro)
    return vserver


def vserver_exists(v_name, v_ip=None, v_port=None, v_type=None, **connection_args):
    """
    Checks if a vserver exists

    CLI Example:

    .. code-block:: bash

        salt '*' netscaler.vserver_exists 'vserverName'
    """
    vserver = _vserver_get(v_name, **connection_args)
    if vserver is None:
        return False
    if v_ip is not None and vserver.get_ipv46() != v_ip:
        return False
    if v_port is not None and vserver.get_port() != v_port:
        return False
    if v_type is not None and vserver.get_servicetype().upper() != v_type.upper():
        return False
    return True


def vserver_add(v_name, v_ip, v_port, v_type, **connection_args):
    """
    Add a new lb vserver

    CLI Example:

    .. code-block:: bash

        salt '*' netscaler.vserver_add 'vserverName' 'vserverIP' 'vserverPort' 'vserverType'
        salt '*' netscaler.vserver_add 'alex.patate.chaude.443' '1.2.3.4' '443' 'SSL'
    """
    ret = True
    if vserver_exists(v_name, **connection_args):
        return False
    nitro = _connect(**connection_args)
    if nitro is None:
        return False
    vserver = NSLBVServer()
    vserver.set_name(v_name)
    vserver.set_ipv46(v_ip)
    vserver.set_port(v_port)
    vserver.set_servicetype(v_type.upper())
    try:
        NSLBVServer.add(nitro, vserver)
    except NSNitroError as error:
        log.debug("netscaler module error - NSLBVServer.add() failed: %s", error)
        ret = False
    _disconnect(nitro)
    return ret


def vserver_delete(v_name, **connection_args):
    """
    Delete a lb vserver

    CLI Example:

    .. code-block:: bash

        salt '*' netscaler.vserver_delete 'vserverName'
    """
    ret = True
    vserver = _vserver_get(v_name, **connection_args)
    if vserver is None:
        return False
    nitro = _connect(**connection_args)
    if nitro is None:
        return False
    try:
        NSLBVServer.delete(nitro, vserver)
    except NSNitroError as error:
        log.debug("netscaler module error - NSVServer.delete() failed: %s", error)
        ret = False
    _disconnect(nitro)
    return ret


def _vserver_servicegroup_get(v_name, sg_name, **connection_args):
    ret = None
    nitro = _connect(**connection_args)
    if nitro is None:
        return None
    vsg = NSLBVServerServiceGroupBinding()
    vsg.set_name(v_name)
    try:
        vsgs = NSLBVServerServiceGroupBinding.get(nitro, vsg)
    except NSNitroError as error:
        log.debug(
            "netscaler module error - NSLBVServerServiceGroupBinding.get() failed: %s",
            error,
        )
        return None
    for vsg in vsgs:
        if vsg.get_servicegroupname() == sg_name:
            ret = vsg
    _disconnect(nitro)
    return ret


def vserver_servicegroup_exists(v_name, sg_name, **connection_args):
    """
    Checks if a servicegroup is tied to a vserver

    CLI Example:

    .. code-block:: bash

        salt '*' netscaler.vserver_servicegroup_exists 'vserverName' 'serviceGroupName'
    """
    return _vserver_servicegroup_get(v_name, sg_name, **connection_args) is not None


def vserver_servicegroup_add(v_name, sg_name, **connection_args):
    """
    Bind a servicegroup to a vserver

    CLI Example:

    .. code-block:: bash

        salt '*' netscaler.vserver_servicegroup_add 'vserverName' 'serviceGroupName'
    """
    ret = True
    if vserver_servicegroup_exists(v_name, sg_name, **connection_args):
        return False
    nitro = _connect(**connection_args)
    if nitro is None:
        return False
    vsg = NSLBVServerServiceGroupBinding()
    vsg.set_name(v_name)
    vsg.set_servicegroupname(sg_name)
    try:
        NSLBVServerServiceGroupBinding.add(nitro, vsg)
    except NSNitroError as error:
        log.debug(
            "netscaler module error - NSLBVServerServiceGroupBinding.add() failed: %s",
            error,
        )
        ret = False
    _disconnect(nitro)
    return ret


def vserver_servicegroup_delete(v_name, sg_name, **connection_args):
    """
    Unbind a servicegroup from a vserver

    CLI Example:

    .. code-block:: bash

        salt '*' netscaler.vserver_servicegroup_delete 'vserverName' 'serviceGroupName'
    """
    ret = True
    if not vserver_servicegroup_exists(v_name, sg_name, **connection_args):
        return False
    nitro = _connect(**connection_args)
    if nitro is None:
        return False
    vsg = NSLBVServerServiceGroupBinding()
    vsg.set_name(v_name)
    vsg.set_servicegroupname(sg_name)
    try:
        NSLBVServerServiceGroupBinding.delete(nitro, vsg)
    except NSNitroError as error:
        log.debug(
            "netscaler module error - NSLBVServerServiceGroupBinding.delete()"
            " failed: %s",
            error,
        )
        ret = False
    _disconnect(nitro)
    return ret


def _vserver_sslcert_get(v_name, sc_name, **connection_args):
    ret = None
    nitro = _connect(**connection_args)
    if nitro is None:
        return None
    sslcert = NSSSLVServerSSLCertKeyBinding()
    sslcert.set_vservername(v_name)
    try:
        sslcerts = NSSSLVServerSSLCertKeyBinding.get(nitro, sslcert)
    except NSNitroError as error:
        log.debug(
            "netscaler module error - NSSSLVServerSSLCertKeyBinding.get() failed: %s",
            error,
        )
        return None
    for sslcert in sslcerts:
        if sslcert.get_certkeyname() == sc_name:
            ret = sslcert
    return ret


def vserver_sslcert_exists(v_name, sc_name, **connection_args):
    """
    Checks if a SSL certificate is tied to a vserver

    CLI Example:

    .. code-block:: bash

        salt '*' netscaler.vserver_sslcert_exists 'vserverName' 'sslCertificateName'
    """
    return _vserver_sslcert_get(v_name, sc_name, **connection_args) is not None


def vserver_sslcert_add(v_name, sc_name, **connection_args):
    """
    Binds a SSL certificate to a vserver

    CLI Example:

    .. code-block:: bash

        salt '*' netscaler.vserver_sslcert_add 'vserverName' 'sslCertificateName'
    """
    ret = True
    if vserver_sslcert_exists(v_name, sc_name, **connection_args):
        return False
    nitro = _connect(**connection_args)
    if nitro is None:
        return False
    sslcert = NSSSLVServerSSLCertKeyBinding()
    sslcert.set_vservername(v_name)
    sslcert.set_certkeyname(sc_name)
    try:
        NSSSLVServerSSLCertKeyBinding.add(nitro, sslcert)
    except NSNitroError as error:
        log.debug(
            "netscaler module error - NSSSLVServerSSLCertKeyBinding.add() failed: %s",
            error,
        )
        ret = False
    _disconnect(nitro)
    return ret


def vserver_sslcert_delete(v_name, sc_name, **connection_args):
    """
    Unbinds a SSL certificate from a vserver

    CLI Example:

    .. code-block:: bash

        salt '*' netscaler.vserver_sslcert_delete 'vserverName' 'sslCertificateName'
    """
    ret = True
    if not vserver_sslcert_exists(v_name, sc_name, **connection_args):
        return False
    nitro = _connect(**connection_args)
    if nitro is None:
        return False
    sslcert = NSSSLVServerSSLCertKeyBinding()
    sslcert.set_vservername(v_name)
    sslcert.set_certkeyname(sc_name)
    try:
        NSSSLVServerSSLCertKeyBinding.delete(nitro, sslcert)
    except NSNitroError as error:
        log.debug(
            "netscaler module error - NSSSLVServerSSLCertKeyBinding.delete()"
            " failed: %s",
            error,
        )
        ret = False
    _disconnect(nitro)
    return ret



import logging

import pytest
import salt.modules.solarisipspkg as solarisipspkg
from tests.support.mock import MagicMock, call, patch

log = logging.getLogger(__name__)


class ListPackages:
    def __init__(self, iteration=0):
        self._iteration = iteration

    def __call__(self):
        pkg_lists = [
            {
                "pkg://solaris/compress/bzip2": (
                    "1.0.6-11.4.0.0.1.14.0:20180814T153143Z"
                ),
                "pkg://solaris/compress/gzip": "1.8-11.4.0.0.1.14.0:20180814T153144Z",
                "pkg://solaris/compress/p7zip": (
                    "16.2.3-11.4.0.0.1.14.0:20180814T153145Z"
                ),
                "pkg://solaris/compress/pbzip2": (
                    "1.1.13-11.4.0.0.1.14.0:20180814T153147Z"
                ),
                "pkg://solaris/compress/unzip": (
                    "6.0.3.23-11.4.0.0.1.14.0:20180814T153150Z"
                ),
            },
            {
                "pkg://solaris/compress/bzip2": (
                    "1.0.6-11.4.0.0.1.14.0:20180814T153143Z"
                ),
                "pkg://solaris/compress/gzip": "1.8-11.4.0.0.1.14.0:20180814T153144Z",
                "pkg://solaris/compress/p7zip": (
                    "16.2.3-11.4.0.0.1.14.0:20180814T153145Z"
                ),
                "pkg://solaris/compress/pbzip2": (
                    "1.1.13-11.4.0.0.1.14.0:20180814T153147Z"
                ),
                "pkg://solaris/compress/unzip": (
                    "6.0.3.23-11.4.0.0.1.14.0:20180814T153150Z"
                ),
                "pkg://solaris/compress/zip": "3.0-11.4.0.0.1.14.0:20180814T153154Z",
            },
        ]
        pkgs = pkg_lists[self._iteration]
        self._iteration = (self._iteration + 1) % 2
        return pkgs


class ListPackagesDict:
    def __init__(self, iteration=0):
        self._iteration = iteration

    def __call__(self):
        pkg_lists = [
            {
                "pkg://solaris/compress/bzip2": (
                    "1.0.6-11.4.0.0.1.14.0:20180814T153143Z"
                ),
                "pkg://solaris/compress/gzip": "1.8-11.4.0.0.1.14.0:20180814T153144Z",
                "pkg://solaris/compress/p7zip": (
                    "16.2.3-11.4.0.0.1.14.0:20180814T153145Z"
                ),
                "pkg://solaris/compress/pbzip2": (
                    "1.1.13-11.4.0.0.1.14.0:20180814T153147Z"
                ),
                "pkg://solaris/compress/unzip": (
                    "6.0.3.23-11.4.0.0.1.14.0:20180814T153150Z"
                ),
            },
            {
                "pkg://solaris/compress/bzip2": (
                    "1.0.6-11.4.0.0.1.14.0:20180814T153143Z"
                ),
                "pkg://solaris/compress/gzip": "1.8-11.4.0.0.1.14.0:20180814T153144Z",
                "pkg://solaris/compress/p7zip": (
                    "16.2.3-11.4.0.0.1.14.0:20180814T153145Z"
                ),
                "pkg://solaris/compress/pbzip2": (
                    "1.1.13-11.4.0.0.1.14.0:20180814T153147Z"
                ),
                "pkg://solaris/compress/unzip": (
                    "6.0.3.23-11.4.0.0.1.14.0:20180814T153150Z"
                ),
                "pkg://solaris/file/tree": "1.7.0-11.4.0.0.1.14.0:20180814T163602Z",
                "pkg://solaris/x11/xclock": "1.0.7-11.4.0.0.1.14.0:20180814T173537Z",
            },
        ]
        pkgs = pkg_lists[self._iteration]
        self._iteration = (self._iteration + 1) % 2
        return pkgs


@pytest.fixture
def configure_loader_modules():

    return {
        solarisipspkg: {
            "__grains__": {
                "osarch": "sparcv9",
                "os_family": "Solaris",
                "osmajorrelease": 11,
                "kernelrelease": 5.11,
            },
        },
    }


def test_list_pkgs():
    """
    Test for listing installed packages.
    """

    def _add_data(data, key, value):
        data[key] = value

    pkg_info_out = [
        "pkg://solaris/compress/bzip2@1.0.6-11.4.0.0.1.14.0:20180814T153143Z        "
        "  i--",
        "pkg://solaris/compress/gzip@1.8-11.4.0.0.1.14.0:20180814T153144Z           "
        "  i--",
        "pkg://solaris/compress/p7zip@16.2.3-11.4.0.0.1.14.0:20180814T153145Z       "
        "  i--",
        "pkg://solaris/compress/pbzip2@1.1.13-11.4.0.0.1.14.0:20180814T153147Z      "
        "  i--",
        "pkg://solaris/compress/unzip@6.0.3.23-11.4.0.0.1.14.0:20180814T153150Z     "
        "  i--",
    ]
    run_stdout_mock = MagicMock(return_value="\n".join(pkg_info_out))
    patches = {
        "cmd.run_stdout": run_stdout_mock,
        "pkg_resource.add_pkg": _add_data,
        "pkg_resource.sort_pkglist": MagicMock(),
        "pkg_resource.stringify": MagicMock(),
    }
    with patch.dict(solarisipspkg.__salt__, patches):
        pkgs = solarisipspkg.list_pkgs()
        assert pkgs == {
            "pkg://solaris/compress/bzip2": "1.0.6-11.4.0.0.1.14.0:20180814T153143Z",
            "pkg://solaris/compress/gzip": "1.8-11.4.0.0.1.14.0:20180814T153144Z",
            "pkg://solaris/compress/p7zip": "16.2.3-11.4.0.0.1.14.0:20180814T153145Z",
            "pkg://solaris/compress/pbzip2": "1.1.13-11.4.0.0.1.14.0:20180814T153147Z",
            "pkg://solaris/compress/unzip": "6.0.3.23-11.4.0.0.1.14.0:20180814T153150Z",
        }
    run_stdout_mock.assert_called_once_with("/bin/pkg list -Hv")


def test_install_single_named_package():
    """
    Test installing a single package
    - a single package pkg://solaris/compress/zip 3.0-11.4.0.0.1.14.0:20180814T153154Z
    """

    install_target = "compress/zip"
    cmd_out = {
        "retcode": 0,
        "stdout": "",
        "stderr": "",
    }
    run_all_mock = MagicMock(return_value=cmd_out)
    patches = {
        "cmd.run_all": run_all_mock,
        "solarisipspkg.is_installed": MagicMock(return_value=False),
        "cmd.retcode": MagicMock(return_value=False),
    }

    with patch.dict(solarisipspkg.__salt__, patches):
        with patch("salt.modules.solarisipspkg.list_pkgs", ListPackages()), patch(
            "salt.modules.solarisipspkg.is_installed", MagicMock(return_value=False)
        ):
            added = solarisipspkg.install(install_target, refresh=False)
            expected = {
                "pkg://solaris/compress/zip": {
                    "new": "3.0-11.4.0.0.1.14.0:20180814T153154Z",
                    "old": "",
                }
            }
            assert added == expected

    expected_calls = [
        call(
            ["pkg", "install", "-v", "--accept", install_target],
            output_loglevel="trace",
        ),
    ]
    run_all_mock.assert_has_calls(expected_calls, any_order=True)
    assert run_all_mock.call_count == 1


def test_install_single_pkg_package():
    """
    Test installing a single package
    - a single package pkg://solaris/compress/zip 3.0-11.4.0.0.1.14.0:20180814T153154Z
    """

    install_target = "pkg://solaris/compress/zip"
    cmd_out = {
        "retcode": 0,
        "stdout": "",
        "stderr": "",
    }
    run_all_mock = MagicMock(return_value=cmd_out)
    patches = {
        "cmd.run_all": run_all_mock,
        "solarisipspkg.is_installed": MagicMock(return_value=False),
        "cmd.retcode": MagicMock(return_value=False),
    }

    with patch.dict(solarisipspkg.__salt__, patches):
        with patch("salt.modules.solarisipspkg.list_pkgs", ListPackages()):
            added = solarisipspkg.install(pkgs=[install_target], refresh=False)
            expected = {
                "pkg://solaris/compress/zip": {
                    "new": "3.0-11.4.0.0.1.14.0:20180814T153154Z",
                    "old": "",
                }
            }
            assert added == expected

    expected_calls = [
        call(
            ["pkg", "install", "-v", "--accept", install_target],
            output_loglevel="trace",
        ),
    ]
    run_all_mock.assert_has_calls(expected_calls, any_order=True)
    assert run_all_mock.call_count == 1


def test_install_dict_pkgs_no_version():
    """
    Test installing a list of packages in a dict with no versions
    """

    install_target = [{"tree": ""}, {"xclock": ""}]
    cmd_out = {
        "retcode": 0,
        "stdout": "",
        "stderr": "",
    }
    run_all_mock = MagicMock(return_value=cmd_out)
    patches = {
        "cmd.run_all": run_all_mock,
        "solarisipspkg.is_installed": MagicMock(return_value=False),
        "cmd.retcode": MagicMock(return_value=False),
    }

    with patch.dict(solarisipspkg.__salt__, patches):
        with patch("salt.modules.solarisipspkg.list_pkgs", ListPackagesDict()), patch(
            "salt.modules.solarisipspkg.is_installed", MagicMock(return_value=False)
        ):
            added = solarisipspkg.install(pkgs=install_target, refresh=False)
            expected = {
                "pkg://solaris/file/tree": {
                    "new": "1.7.0-11.4.0.0.1.14.0:20180814T163602Z",
                    "old": "",
                },
                "pkg://solaris/x11/xclock": {
                    "new": "1.0.7-11.4.0.0.1.14.0:20180814T173537Z",
                    "old": "",
                },
            }
            assert added == expected

    list_first = tuple(install_target[0].keys())[0]
    list_second = tuple(install_target[1].keys())[0]
    expected_calls = [
        call(
            ["pkg", "install", "-v", "--accept", list_first, list_second],
            output_loglevel="trace",
        ),
    ]
    run_all_mock.assert_has_calls(expected_calls, any_order=True)
    assert run_all_mock.call_count == 1


def test_install_dict_pkgs_with_version():
    """
    Test installing a list of packages in a dict with versions
    """

    install_target = [
        {"tree": "1.7.0-11.4.0.0.1.14.0:20180814T163602Z"},
        {"xclock": "1.0.7-11.4.0.0.1.14.0:20180814T173537Z"},
    ]
    cmd_out = {
        "retcode": 0,
        "stdout": "",
        "stderr": "",
    }
    run_all_mock = MagicMock(return_value=cmd_out)
    patches = {
        "cmd.run_all": run_all_mock,
        "solarisipspkg.is_installed": MagicMock(return_value=False),
        "cmd.retcode": MagicMock(return_value=False),
    }

    with patch.dict(solarisipspkg.__salt__, patches):
        with patch("salt.modules.solarisipspkg.list_pkgs", ListPackagesDict()), patch(
            "salt.modules.solarisipspkg.is_installed", MagicMock(return_value=False)
        ):
            added = solarisipspkg.install(pkgs=install_target, refresh=False)
            expected = {
                "pkg://solaris/file/tree": {
                    "new": "1.7.0-11.4.0.0.1.14.0:20180814T163602Z",
                    "old": "",
                },
                "pkg://solaris/x11/xclock": {
                    "new": "1.0.7-11.4.0.0.1.14.0:20180814T173537Z",
                    "old": "",
                },
            }
            assert added == expected

    list_first = (
        tuple(install_target[0].keys())[0] + "@1.7.0-11.4.0.0.1.14.0:20180814T163602Z"
    )
    list_second = (
        tuple(install_target[1].keys())[0] + "@1.0.7-11.4.0.0.1.14.0:20180814T173537Z"
    )
    expected_calls = [
        call(
            ["pkg", "install", "-v", "--accept", list_first, list_second],
            output_loglevel="trace",
        ),
    ]
    run_all_mock.assert_has_calls(expected_calls, any_order=True)
    assert run_all_mock.call_count == 1


def test_install_already_installed_single_pkg():
    """
    Test installing a package that is already installed
    """
    result = None
    expected_result = {}
    install_target = "compress/zip"

    patches = {
        "solarisipspkg.is_installed": MagicMock(return_value=True),
        "cmd.retcode": MagicMock(return_value=False),
    }

    with patch.dict(solarisipspkg.__salt__, patches):
        result = solarisipspkg.install(install_target)
        assert result == expected_result


def test_get_fmri_single_named():
    """
    Test getting the frmi for a specific package
    - a single package pkg://solaris/compress/zip 3.0-11.4.0.0.1.14.0:20180814T153154Z
    """

    install_target = "compress/zip"
    pkg_info_out = [
        "pkg://solaris/compress/zip@3.0-11.4.0.0.1.14.0:20180814T153154Z            "
        "  i--",
    ]

    run_stdout_mock = MagicMock(return_value="\n".join(pkg_info_out))
    patches = {
        "cmd.run_stdout": run_stdout_mock,
    }

    with patch.dict(solarisipspkg.__salt__, patches):
        with patch("salt.modules.solarisipspkg.list_pkgs", ListPackages(1)):
            added = solarisipspkg.get_fmri(install_target, refresh=False)
            expected = ["pkg://solaris/compress/zip"]
            assert added == expected

    expected_calls = [
        call(
            ["/bin/pkg", "list", "-aHv", install_target],
        ),
    ]
    run_stdout_mock.assert_has_calls(expected_calls, any_order=True)
    assert run_stdout_mock.call_count == 1


def test_get_fmri_single_pkg():
    """
    Test getting the frmi for a specific package
    - a single package pkg://solaris/compress/zip 3.0-11.4.0.0.1.14.0:20180814T153154Z
    """

    install_target = "pkg://solaris/compress/zip"
    result = solarisipspkg.get_fmri(install_target, refresh=False)
    assert result == install_target


def test_remove_single_named_package():
    """
    Test removing a single specific named package
    - a single package compress/zip
    """

    install_target = "compress/zip"
    cmd_out = {
        "retcode": 0,
        "stdout": "",
        "stderr": "",
    }
    run_all_mock = MagicMock(return_value=cmd_out)
    patches = {
        "cmd.run_all": run_all_mock,
        "solarisipspkg.is_installed": MagicMock(return_value=False),
        "cmd.retcode": MagicMock(return_value=False),
    }

    with patch.dict(solarisipspkg.__salt__, patches):
        with patch("salt.modules.solarisipspkg.list_pkgs", ListPackages(1)):
            added = solarisipspkg.remove(install_target, refresh=False)
            expected = {
                "pkg://solaris/compress/zip": {
                    "new": "",
                    "old": "3.0-11.4.0.0.1.14.0:20180814T153154Z",
                }
            }
            assert added == expected

    expected_calls = [
        call(
            ["/bin/pkg", "uninstall", "-v", install_target],
            output_loglevel="trace",
        ),
    ]
    run_all_mock.assert_has_calls(expected_calls, any_order=True)
    assert run_all_mock.call_count == 1


def test_remove_listed_pkgs():
    """
    Test removing a list of packages
    """
    install_target = [{"tree": ""}, {"xclock": ""}]
    cmd_out = {
        "retcode": 0,
        "stdout": "",
        "stderr": "",
    }
    run_all_mock = MagicMock(return_value=cmd_out)
    patches = {
        "cmd.run_all": run_all_mock,
        "solarisipspkg.is_installed": MagicMock(return_value=False),
        "cmd.retcode": MagicMock(return_value=False),
    }

    with patch.dict(solarisipspkg.__salt__, patches):
        with patch("salt.modules.solarisipspkg.list_pkgs", ListPackagesDict(1)), patch(
            "salt.modules.solarisipspkg.is_installed", MagicMock(return_value=False)
        ):
            added = solarisipspkg.remove(pkgs=install_target, refresh=False)
            expected = {
                "pkg://solaris/file/tree": {
                    "new": "",
                    "old": "1.7.0-11.4.0.0.1.14.0:20180814T163602Z",
                },
                "pkg://solaris/x11/xclock": {
                    "new": "",
                    "old": "1.0.7-11.4.0.0.1.14.0:20180814T173537Z",
                },
            }
            assert added == expected

    expected_calls = [
        call(
            ["/bin/pkg", "uninstall", "-v", install_target[0], install_target[1]],
            output_loglevel="trace",
        ),
    ]
    run_all_mock.assert_has_calls(expected_calls, any_order=True)
    assert run_all_mock.call_count == 1



"""
Validate the mac-keychain module
"""

import os

import pytest
from salt.exceptions import CommandExecutionError
from tests.support.case import ModuleCase
from tests.support.helpers import runs_on
from tests.support.runtests import RUNTIME_VARS


@pytest.mark.destructive_test
@runs_on(kernel="Darwin")
@pytest.mark.skip_if_not_root
class MacKeychainModuleTest(ModuleCase):
    """
    Integration tests for the mac_keychain module
    """

    @classmethod
    def setUpClass(cls):
        cls.cert = os.path.join(
            RUNTIME_VARS.FILES, "file", "base", "certs", "salttest.p12"
        )
        cls.cert_alias = "Salt Test"
        cls.passwd = "salttest"

    def tearDown(self):
        """
        Clean up after tests
        """
        # Remove the salttest cert, if left over.
        certs_list = self.run_function("keychain.list_certs")
        if self.cert_alias in certs_list:
            self.run_function("keychain.uninstall", [self.cert_alias])

    @pytest.mark.slow_test
    def test_mac_keychain_install(self):
        """
        Tests that attempts to install a certificate
        """
        install_cert = self.run_function("keychain.install", [self.cert, self.passwd])
        self.assertTrue(install_cert)

        # check to ensure the cert was installed
        certs_list = self.run_function("keychain.list_certs")
        self.assertIn(self.cert_alias, certs_list)

    @pytest.mark.slow_test
    def test_mac_keychain_uninstall(self):
        """
        Tests that attempts to uninstall a certificate
        """
        self.run_function("keychain.install", [self.cert, self.passwd])
        certs_list = self.run_function("keychain.list_certs")

        if self.cert_alias not in certs_list:
            self.run_function("keychain.uninstall", [self.cert_alias])
            self.skipTest("Failed to install keychain")

        # uninstall cert
        self.run_function("keychain.uninstall", [self.cert_alias])
        certs_list = self.run_function("keychain.list_certs")

        # check to ensure the cert was uninstalled
        try:
            self.assertNotIn(self.cert_alias, str(certs_list))
        except CommandExecutionError:
            self.run_function("keychain.uninstall", [self.cert_alias])

    @pytest.mark.slow_test
    def test_mac_keychain_get_friendly_name(self):
        """
        Test that attempts to get friendly name of a cert
        """
        self.run_function("keychain.install", [self.cert, self.passwd])
        certs_list = self.run_function("keychain.list_certs")
        if self.cert_alias not in certs_list:
            self.run_function("keychain.uninstall", [self.cert_alias])
            self.skipTest("Failed to install keychain")

        get_name = self.run_function(
            "keychain.get_friendly_name", [self.cert, self.passwd]
        )
        self.assertEqual(get_name, self.cert_alias)

    @pytest.mark.slow_test
    def test_mac_keychain_get_default_keychain(self):
        """
        Test that attempts to get the default keychain
        """
        salt_get_keychain = self.run_function("keychain.get_default_keychain")
        sys_get_keychain = self.run_function(
            "cmd.run", ["security default-keychain -d user"]
        )
        self.assertEqual(salt_get_keychain, sys_get_keychain)

    def test_mac_keychain_list_certs(self):
        """
        Test that attempts to list certs
        """
        cert_default = "com.apple.systemdefault"
        certs = self.run_function("keychain.list_certs")
        self.assertIn(cert_default, certs)



import textwrap

import salt.modules.parallels as parallels
from salt.exceptions import CommandExecutionError, SaltInvocationError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class ParallelsTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test parallels desktop execution module functions
    """

    def setup_loader_modules(self):
        return {parallels: {}}

    def test__normalize_args(self):
        """
        Test parallels._normalize_args
        """

        def _validate_ret(ret):
            """
            Assert that the returned data is a list of strings
            """
            self.assertTrue(isinstance(ret, list))
            for arg in ret:
                self.assertTrue(isinstance(arg, str))

        # Validate string arguments
        str_args = (
            "electrolytes --aqueous --anion hydroxide --cation=ammonium free radicals"
            " -- hydrogen"
        )
        _validate_ret(parallels._normalize_args(str_args))

        # Validate list arguments
        list_args = " ".join(str_args)
        _validate_ret(parallels._normalize_args(list_args))

        # Validate tuple arguments
        tuple_args = tuple(list_args)
        _validate_ret(parallels._normalize_args(tuple_args))

        # Validate dictionary arguments
        other_args = {"anion": "hydroxide", "cation": "ammonium"}
        _validate_ret(parallels._normalize_args(other_args))

    def test__find_guids(self):
        """
        Test parallels._find_guids
        """
        guid_str = textwrap.dedent(
            """
            PARENT_SNAPSHOT_ID                      SNAPSHOT_ID
                                                    {a5b8999f-5d95-4aff-82de-e515b0101b66}
            {a5b8999f-5d95-4aff-82de-e515b0101b66} *{a7345be5-ab66-478c-946e-a6c2caf14909}
        """
        )
        guids = [
            "a5b8999f-5d95-4aff-82de-e515b0101b66",
            "a7345be5-ab66-478c-946e-a6c2caf14909",
        ]

        self.assertEqual(parallels._find_guids(guid_str), guids)

    def test_prlsrvctl(self):
        """
        Test parallels.prlsrvctl
        """
        runas = "macdev"

        # Test missing prlsrvctl binary
        with patch("salt.utils.path.which", MagicMock(return_value=False)):
            with self.assertRaises(CommandExecutionError):
                parallels.prlsrvctl("info", runas=runas)

        # Simulate the existence of prlsrvctl
        with patch(
            "salt.utils.path.which", MagicMock(return_value="/usr/bin/prlsrvctl")
        ):

            # Validate 'prlsrvctl info'
            info_cmd = ["prlsrvctl", "info"]
            info_fcn = MagicMock()
            with patch.dict(parallels.__salt__, {"cmd.run": info_fcn}):
                parallels.prlsrvctl("info", runas=runas)
                info_fcn.assert_called_once_with(info_cmd, runas=runas)

            # Validate 'prlsrvctl usb list'
            usb_cmd = ["prlsrvctl", "usb", "list"]
            usb_fcn = MagicMock()
            with patch.dict(parallels.__salt__, {"cmd.run": usb_fcn}):
                parallels.prlsrvctl("usb", "list", runas=runas)
                usb_fcn.assert_called_once_with(usb_cmd, runas=runas)

            # Validate 'prlsrvctl set "--mem-limit auto"'
            set_cmd = ["prlsrvctl", "set", "--mem-limit", "auto"]
            set_fcn = MagicMock()
            with patch.dict(parallels.__salt__, {"cmd.run": set_fcn}):
                parallels.prlsrvctl("set", "--mem-limit auto", runas=runas)
                set_fcn.assert_called_once_with(set_cmd, runas=runas)

    def test_prlctl(self):
        """
        Test parallels.prlctl
        """
        runas = "macdev"

        # Test missing prlctl binary
        with patch("salt.utils.path.which", MagicMock(return_value=False)):
            with self.assertRaises(CommandExecutionError):
                parallels.prlctl("info", runas=runas)

        # Simulate the existence of prlctl
        with patch("salt.utils.path.which", MagicMock(return_value="/usr/bin/prlctl")):

            # Validate 'prlctl user list'
            user_cmd = ["prlctl", "user", "list"]
            user_fcn = MagicMock()
            with patch.dict(parallels.__salt__, {"cmd.run": user_fcn}):
                parallels.prlctl("user", "list", runas=runas)
                user_fcn.assert_called_once_with(user_cmd, runas=runas)

            # Validate 'prlctl exec "macvm uname"'
            exec_cmd = ["prlctl", "exec", "macvm", "uname"]
            exec_fcn = MagicMock()
            with patch.dict(parallels.__salt__, {"cmd.run": exec_fcn}):
                parallels.prlctl("exec", "macvm uname", runas=runas)
                exec_fcn.assert_called_once_with(exec_cmd, runas=runas)

            # Validate 'prlctl capture "macvm --file macvm.display.png"'
            cap_cmd = ["prlctl", "capture", "macvm", "--file", "macvm.display.png"]
            cap_fcn = MagicMock()
            with patch.dict(parallels.__salt__, {"cmd.run": cap_fcn}):
                parallels.prlctl(
                    "capture", "macvm --file macvm.display.png", runas=runas
                )
                cap_fcn.assert_called_once_with(cap_cmd, runas=runas)

    def test_list_vms(self):
        """
        Test parallels.list_vms
        """
        runas = "macdev"

        # Validate a simple list
        mock_plain = MagicMock()
        with patch.object(parallels, "prlctl", mock_plain):
            parallels.list_vms(runas=runas)
            mock_plain.assert_called_once_with("list", [], runas=runas)

        # Validate listing a single VM
        mock_name = MagicMock()
        with patch.object(parallels, "prlctl", mock_name):
            parallels.list_vms(name="macvm", runas=runas)
            mock_name.assert_called_once_with("list", ["macvm"], runas=runas)

        # Validate listing templates
        mock_templ = MagicMock()
        with patch.object(parallels, "prlctl", mock_templ):
            parallels.list_vms(template=True, runas=runas)
            mock_templ.assert_called_once_with("list", ["--template"], runas=runas)

        # Validate listing extra info
        mock_info = MagicMock()
        with patch.object(parallels, "prlctl", mock_info):
            parallels.list_vms(info=True, runas=runas)
            mock_info.assert_called_once_with("list", ["--info"], runas=runas)

        # Validate listing with extra options
        mock_complex = MagicMock()
        with patch.object(parallels, "prlctl", mock_complex):
            parallels.list_vms(args=" -o uuid,status", all=True, runas=runas)
            mock_complex.assert_called_once_with(
                "list", ["-o", "uuid,status", "--all"], runas=runas
            )

    def test_clone(self):
        """
        Test parallels.clone
        """
        name = "macvm"
        runas = "macdev"

        # Validate clone
        mock_clone = MagicMock()
        with patch.object(parallels, "prlctl", mock_clone):
            parallels.clone(name, "macvm_new", runas=runas)
            mock_clone.assert_called_once_with(
                "clone", [name, "--name", "macvm_new"], runas=runas
            )

        # Validate linked clone
        mock_linked = MagicMock()
        with patch.object(parallels, "prlctl", mock_linked):
            parallels.clone(name, "macvm_link", linked=True, runas=runas)
            mock_linked.assert_called_once_with(
                "clone", [name, "--name", "macvm_link", "--linked"], runas=runas
            )

        # Validate template clone
        mock_template = MagicMock()
        with patch.object(parallels, "prlctl", mock_template):
            parallels.clone(name, "macvm_templ", template=True, runas=runas)
            mock_template.assert_called_once_with(
                "clone", [name, "--name", "macvm_templ", "--template"], runas=runas
            )

    def test_delete(self):
        """
        Test parallels.delete
        """
        name = "macvm"
        runas = "macdev"

        # Validate delete
        mock_delete = MagicMock()
        with patch.object(parallels, "prlctl", mock_delete):
            parallels.delete(name, runas=runas)
            mock_delete.assert_called_once_with("delete", name, runas=runas)

    def test_exists(self):
        """
        Test parallels.exists
        """
        name = "macvm"
        runas = "macdev"

        # Validate exists
        mock_list = MagicMock(return_value="Name: {}\nState: running".format(name))
        with patch.object(parallels, "list_vms", mock_list):
            self.assertTrue(parallels.exists(name, runas=runas))

        # Validate not exists
        mock_list = MagicMock(return_value="Name: {}\nState: running".format(name))
        with patch.object(parallels, "list_vms", mock_list):
            self.assertFalse(parallels.exists("winvm", runas=runas))

    def test_start(self):
        """
        Test parallels.start
        """
        name = "macvm"
        runas = "macdev"

        mock_start = MagicMock()
        with patch.object(parallels, "prlctl", mock_start):
            parallels.start(name, runas=runas)
            mock_start.assert_called_once_with("start", name, runas=runas)

    def test_stop(self):
        """
        Test parallels.stop
        """
        name = "macvm"
        runas = "macdev"

        # Validate stop
        mock_stop = MagicMock()
        with patch.object(parallels, "prlctl", mock_stop):
            parallels.stop(name, runas=runas)
            mock_stop.assert_called_once_with("stop", [name], runas=runas)

        # Validate immediate stop
        mock_kill = MagicMock()
        with patch.object(parallels, "prlctl", mock_kill):
            parallels.stop(name, kill=True, runas=runas)
            mock_kill.assert_called_once_with("stop", [name, "--kill"], runas=runas)

    def test_restart(self):
        """
        Test parallels.restart
        """
        name = "macvm"
        runas = "macdev"

        mock_start = MagicMock()
        with patch.object(parallels, "prlctl", mock_start):
            parallels.restart(name, runas=runas)
            mock_start.assert_called_once_with("restart", name, runas=runas)

    def test_reset(self):
        """
        Test parallels.reset
        """
        name = "macvm"
        runas = "macdev"

        mock_start = MagicMock()
        with patch.object(parallels, "prlctl", mock_start):
            parallels.reset(name, runas=runas)
            mock_start.assert_called_once_with("reset", name, runas=runas)

    def test_status(self):
        """
        Test parallels.status
        """
        name = "macvm"
        runas = "macdev"

        mock_start = MagicMock()
        with patch.object(parallels, "prlctl", mock_start):
            parallels.status(name, runas=runas)
            mock_start.assert_called_once_with("status", name, runas=runas)

    def test_exec_(self):
        """
        Test parallels.exec_
        """
        name = "macvm"
        runas = "macdev"

        mock_start = MagicMock()
        with patch.object(parallels, "prlctl", mock_start):
            parallels.exec_(name, "find /etc/paths.d", runas=runas)
            mock_start.assert_called_once_with(
                "exec", [name, "find", "/etc/paths.d"], runas=runas
            )

    def test_snapshot_id_to_name(self):
        """
        Test parallels.snapshot_id_to_name
        """
        name = "macvm"
        snap_id = "a5b8999f-5d95-4aff-82de-e515b0101b66"

        # Invalid GUID raises error
        self.assertRaises(
            SaltInvocationError, parallels.snapshot_id_to_name, name, "{8-4-4-4-12}"
        )

        # Empty return from prlctl raises error (name/snap_id mismatch?)
        mock_no_data = MagicMock(return_value="")
        with patch.object(parallels, "prlctl", mock_no_data):
            self.assertRaises(
                SaltInvocationError, parallels.snapshot_id_to_name, name, snap_id
            )

        # Data returned from prlctl is invalid YAML
        mock_invalid_data = MagicMock(return_value="[string theory is falsifiable}")
        with patch.object(parallels, "prlctl", mock_invalid_data):
            snap_name = parallels.snapshot_id_to_name(name, snap_id)
            self.assertEqual(snap_name, "")

        # Data returned from prlctl does not render as a dictionary
        mock_unknown_data = MagicMock(return_value="['sfermions', 'bosinos']")
        with patch.object(parallels, "prlctl", mock_unknown_data):
            snap_name = parallels.snapshot_id_to_name(name, snap_id)
            self.assertEqual(snap_name, "")

        # Snapshot is unnamed
        mock_no_name = MagicMock(return_value="Name:")
        with patch.object(parallels, "prlctl", mock_no_name):
            snap_name = parallels.snapshot_id_to_name(name, snap_id)
            self.assertEqual(snap_name, "")

        # If strict, then raise an error when name is not found
        mock_no_name = MagicMock(return_value="Name:")
        with patch.object(parallels, "prlctl", mock_no_name):
            self.assertRaises(
                SaltInvocationError,
                parallels.snapshot_id_to_name,
                name,
                snap_id,
                strict=True,
            )

        # Return name when found
        mock_yes_name = MagicMock(return_value="Name: top")
        with patch.object(parallels, "prlctl", mock_yes_name):
            snap_name = parallels.snapshot_id_to_name(name, snap_id)
            self.assertEqual(snap_name, "top")

    def test_snapshot_name_to_id(self):
        """
        Test parallels.snapshot_name_to_id
        """
        name = "macvm"
        snap_ids = [
            "a5b8999f-5d95-4aff-82de-e515b0101b66",
            "a7345be5-ab66-478c-946e-a6c2caf14909",
        ]
        snap_id = snap_ids[0]
        guid_str = textwrap.dedent(
            """
            PARENT_SNAPSHOT_ID                      SNAPSHOT_ID
                                                    {a5b8999f-5d95-4aff-82de-e515b0101b66}
            {a5b8999f-5d95-4aff-82de-e515b0101b66} *{a7345be5-ab66-478c-946e-a6c2caf14909}
        """
        )
        mock_guids = MagicMock(return_value=guid_str)

        # Raise error when no IDs found for snapshot name
        with patch.object(parallels, "prlctl", mock_guids):
            mock_no_names = MagicMock(return_value=[])
            with patch.object(parallels, "snapshot_id_to_name", mock_no_names):
                self.assertRaises(
                    SaltInvocationError, parallels.snapshot_name_to_id, name, "graviton"
                )

        # Validate singly-valued name
        with patch.object(parallels, "prlctl", mock_guids):
            mock_one_name = MagicMock(side_effect=["", "ν_e"])
            with patch.object(parallels, "snapshot_id_to_name", mock_one_name):
                self.assertEqual(
                    parallels.snapshot_name_to_id(name, "ν_e"), snap_ids[1]
                )

        # Validate multiply-valued name
        with patch.object(parallels, "prlctl", mock_guids):
            mock_many_names = MagicMock(side_effect=["J/Ψ", "J/Ψ"])
            with patch.object(parallels, "snapshot_id_to_name", mock_many_names):
                self.assertEqual(
                    sorted(parallels.snapshot_name_to_id(name, "J/Ψ")),
                    sorted(snap_ids),
                )

        # Raise error for multiply-valued name
        with patch.object(parallels, "prlctl", mock_guids):
            mock_many_names = MagicMock(side_effect=["J/Ψ", "J/Ψ"])
            with patch.object(parallels, "snapshot_id_to_name", mock_many_names):
                self.assertRaises(
                    SaltInvocationError,
                    parallels.snapshot_name_to_id,
                    name,
                    "J/Ψ",
                    strict=True,
                )

    def test__validate_snap_name(self):
        """
        Test parallels._validate_snap_name
        """
        name = "macvm"
        snap_id = "a5b8999f-5d95-4aff-82de-e515b0101b66"

        # Validate a GUID passthrough
        self.assertEqual(parallels._validate_snap_name(name, snap_id), snap_id)

        # Validate an unicode name
        mock_snap_symb = MagicMock(return_value=snap_id)
        with patch.object(parallels, "snapshot_name_to_id", mock_snap_symb):
            self.assertEqual(parallels._validate_snap_name(name, "π"), snap_id)
            mock_snap_symb.assert_called_once_with(name, "π", strict=True, runas=None)

        # Validate an ascii name
        mock_snap_name = MagicMock(return_value=snap_id)
        with patch.object(parallels, "snapshot_name_to_id", mock_snap_name):
            self.assertEqual(parallels._validate_snap_name(name, "pion"), snap_id)
            mock_snap_name.assert_called_once_with(
                name, "pion", strict=True, runas=None
            )

        # Validate a numerical name
        mock_snap_numb = MagicMock(return_value=snap_id)
        with patch.object(parallels, "snapshot_name_to_id", mock_snap_numb):
            self.assertEqual(parallels._validate_snap_name(name, "3.14159"), snap_id)
            mock_snap_numb.assert_called_once_with(
                name, "3.14159", strict=True, runas=None
            )

        # Validate not strict (think neutrino oscillation)
        mock_snap_non_strict = MagicMock(return_value=snap_id)
        with patch.object(parallels, "snapshot_name_to_id", mock_snap_non_strict):
            self.assertEqual(
                parallels._validate_snap_name(name, "e_ν", strict=False), snap_id
            )
            mock_snap_non_strict.assert_called_once_with(
                name, "e_ν", strict=False, runas=None
            )

    def test_list_snapshots(self):
        """
        Test parallels.list_snapshots
        """
        name = "macvm"
        guid_str = textwrap.dedent(
            """
            PARENT_SNAPSHOT_ID                      SNAPSHOT_ID
                                                    {a5b8999f-5d95-4aff-82de-e515b0101b66}
            {a5b8999f-5d95-4aff-82de-e515b0101b66} *{a7345be5-ab66-478c-946e-a6c2caf14909}
            {a5b8999f-5d95-4aff-82de-e515b0101b66}  {5da9faef-cb0e-466d-9b41-e5571b62ac2a}
        """
        )

        # Validate listing all snapshots for the VM
        mock_prlctl = MagicMock(return_value=guid_str)
        with patch.object(parallels, "prlctl", mock_prlctl):
            parallels.list_snapshots(name)
            mock_prlctl.assert_called_once_with("snapshot-list", [name], runas=None)

        # Validate listing all snapshots in tree mode
        mock_prlctl = MagicMock(return_value=guid_str)
        with patch.object(parallels, "prlctl", mock_prlctl):
            parallels.list_snapshots(name, tree=True)
            mock_prlctl.assert_called_once_with(
                "snapshot-list", [name, "--tree"], runas=None
            )

        # Validate listing a single snapshot
        snap_name = "muon"
        mock_snap_name = MagicMock(return_value=snap_name)
        with patch.object(parallels, "_validate_snap_name", mock_snap_name):
            mock_prlctl = MagicMock(return_value=guid_str)
            with patch.object(parallels, "prlctl", mock_prlctl):
                parallels.list_snapshots(name, snap_name)
                mock_prlctl.assert_called_once_with(
                    "snapshot-list", [name, "--id", snap_name], runas=None
                )

        # Validate listing snapshot ID and name pairs
        snap_names = ["electron", "muon", "tauon"]
        mock_snap_name = MagicMock(side_effect=snap_names)
        with patch.object(parallels, "snapshot_id_to_name", mock_snap_name):
            mock_prlctl = MagicMock(return_value=guid_str)
            with patch.object(parallels, "prlctl", mock_prlctl):
                ret = parallels.list_snapshots(name, names=True)
                for snap_name in snap_names:
                    self.assertIn(snap_name, ret)
                mock_prlctl.assert_called_once_with("snapshot-list", [name], runas=None)

    def test_snapshot(self):
        """
        Test parallels.snapshot
        """
        name = "macvm"

        # Validate creating a snapshot
        mock_snap = MagicMock(return_value="")
        with patch.object(parallels, "prlctl", mock_snap):
            parallels.snapshot(name)
            mock_snap.assert_called_once_with("snapshot", [name], runas=None)

        # Validate creating a snapshot with a name
        snap_name = "h_0"
        mock_snap_name = MagicMock(return_value="")
        with patch.object(parallels, "prlctl", mock_snap_name):
            parallels.snapshot(name, snap_name)
            mock_snap_name.assert_called_once_with(
                "snapshot", [name, "--name", snap_name], runas=None
            )

        # Validate creating a snapshot with a name and a description
        snap_name = "h_0"
        snap_desc = textwrap.dedent(
            "The ground state particle of the higgs multiplet family of bosons"
        )
        mock_snap_name = MagicMock(return_value="")
        with patch.object(parallels, "prlctl", mock_snap_name):
            parallels.snapshot(name, snap_name, snap_desc)
            mock_snap_name.assert_called_once_with(
                "snapshot",
                [name, "--name", snap_name, "--description", snap_desc],
                runas=None,
            )

    def test_delete_snapshot(self):
        """
        Test parallels.delete_snapshot
        """
        delete_message = (
            "Delete the snapshot...\nThe snapshot has been successfully deleted."
        )

        # Validate single ID
        name = "macvm"
        snap_name = "kaon"
        snap_id = "c2eab062-a635-4ccd-b9ae-998370f898b5"

        mock_snap_name = MagicMock(return_value=snap_id)
        with patch.object(parallels, "_validate_snap_name", mock_snap_name):
            mock_delete = MagicMock(return_value=delete_message)
            with patch.object(parallels, "prlctl", mock_delete):
                ret = parallels.delete_snapshot(name, snap_name)
                self.assertEqual(ret, delete_message)
                mock_delete.assert_called_once_with(
                    "snapshot-delete", [name, "--id", snap_id], runas=None
                )

        # Validate multiple IDs
        name = "macvm"
        snap_name = "higgs doublet"
        snap_ids = [
            "c2eab062-a635-4ccd-b9ae-998370f898b5",
            "8aca07c5-a0e1-4dcb-ba75-cb154d46d516",
        ]

        mock_snap_ids = MagicMock(return_value=snap_ids)
        with patch.object(parallels, "_validate_snap_name", mock_snap_ids):
            mock_delete = MagicMock(return_value=delete_message)
            with patch.object(parallels, "prlctl", mock_delete):
                ret = parallels.delete_snapshot(name, snap_name, all=True)
                mock_ret = {snap_ids[0]: delete_message, snap_ids[1]: delete_message}
                self.assertDictEqual(ret, mock_ret)
                mock_delete.assert_any_call(
                    "snapshot-delete", [name, "--id", snap_ids[0]], runas=None
                )
                mock_delete.assert_any_call(
                    "snapshot-delete", [name, "--id", snap_ids[1]], runas=None
                )

    def test_revert_snapshot(self):
        """
        Test parallels.revert_snapshot
        """
        name = "macvm"
        snap_name = "k-bar"
        snap_id = "c2eab062-a635-4ccd-b9ae-998370f898b5"

        mock_snap_name = MagicMock(return_value=snap_id)
        with patch.object(parallels, "_validate_snap_name", mock_snap_name):
            mock_delete = MagicMock(return_value="")
            with patch.object(parallels, "prlctl", mock_delete):
                parallels.revert_snapshot(name, snap_name)
                mock_delete.assert_called_once_with(
                    "snapshot-switch", [name, "--id", snap_id], runas=None
                )



"""
    :codeauthor: Eric Vz <eric@base10.org>
"""


import pytest
import salt.modules.pacmanpkg as pacman
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {pacman: {}}


def test_list_pkgs():
    """
    Test if it list the packages currently installed in a dict
    """
    cmdmock = MagicMock(return_value="A 1.0\nB 2.0")
    sortmock = MagicMock()
    stringifymock = MagicMock()
    mock_ret = {"A": ["1.0"], "B": ["2.0"]}
    with patch.dict(
        pacman.__salt__,
        {
            "cmd.run": cmdmock,
            "pkg_resource.add_pkg": lambda pkgs, name, version: pkgs.setdefault(
                name, []
            ).append(version),
            "pkg_resource.sort_pkglist": sortmock,
            "pkg_resource.stringify": stringifymock,
        },
    ):
        assert pacman.list_pkgs() == mock_ret

    sortmock.assert_called_with(mock_ret)
    stringifymock.assert_called_with(mock_ret)


def test_list_pkgs_as_list():
    """
    Test if it lists the packages currently installed in a dict
    """
    cmdmock = MagicMock(return_value="A 1.0\nB 2.0")
    sortmock = MagicMock()
    stringifymock = MagicMock()
    mock_ret = {"A": ["1.0"], "B": ["2.0"]}
    with patch.dict(
        pacman.__salt__,
        {
            "cmd.run": cmdmock,
            "pkg_resource.add_pkg": lambda pkgs, name, version: pkgs.setdefault(
                name, []
            ).append(version),
            "pkg_resource.sort_pkglist": sortmock,
            "pkg_resource.stringify": stringifymock,
        },
    ):
        assert pacman.list_pkgs(True) == mock_ret

    sortmock.assert_called_with(mock_ret)
    assert stringifymock.call_count == 0


def test_list_pkgs_no_context():
    """
    Test if it list the packages currently installed in a dict
    """
    cmdmock = MagicMock(return_value="A 1.0\nB 2.0")
    sortmock = MagicMock()
    stringifymock = MagicMock()
    mock_ret = {"A": ["1.0"], "B": ["2.0"]}
    with patch.dict(
        pacman.__salt__,
        {
            "cmd.run": cmdmock,
            "pkg_resource.add_pkg": lambda pkgs, name, version: pkgs.setdefault(
                name, []
            ).append(version),
            "pkg_resource.sort_pkglist": sortmock,
            "pkg_resource.stringify": stringifymock,
        },
    ), patch.object(pacman, "_list_pkgs_from_context") as list_pkgs_context_mock:
        assert pacman.list_pkgs() == mock_ret

        pkgs = pacman.list_pkgs(versions_as_list=True, use_context=False)
        list_pkgs_context_mock.assert_not_called()
        list_pkgs_context_mock.reset_mock()

        pkgs = pacman.list_pkgs(versions_as_list=True, use_context=False)
        list_pkgs_context_mock.assert_not_called()
        list_pkgs_context_mock.reset_mock()


def test_group_list():
    """
    Test if it lists the available groups
    """

    def cmdlist(cmd, **kwargs):
        """
        Handle several different commands being run
        """
        if cmd == ["pacman", "-Sgg"]:
            return (
                "group-a pkg1\ngroup-a pkg2\ngroup-f pkg9\ngroup-c pkg3\ngroup-b pkg4"
            )
        elif cmd == ["pacman", "-Qg"]:
            return "group-a pkg1\ngroup-b pkg4"
        else:
            return "Untested command ({}, {})!".format(cmd, kwargs)

    cmdmock = MagicMock(side_effect=cmdlist)

    sortmock = MagicMock()
    with patch.dict(
        pacman.__salt__, {"cmd.run": cmdmock, "pkg_resource.sort_pkglist": sortmock}
    ):
        assert pacman.group_list() == {
            "available": ["group-c", "group-f"],
            "installed": ["group-b"],
            "partially_installed": ["group-a"],
        }


def test_group_info():
    """
    Test if it shows the packages in a group
    """

    def cmdlist(cmd, **kwargs):
        """
        Handle several different commands being run
        """
        if cmd == ["pacman", "-Sgg", "testgroup"]:
            return "testgroup pkg1\ntestgroup pkg2"
        else:
            return "Untested command ({}, {})!".format(cmd, kwargs)

    cmdmock = MagicMock(side_effect=cmdlist)

    sortmock = MagicMock()
    with patch.dict(
        pacman.__salt__, {"cmd.run": cmdmock, "pkg_resource.sort_pkglist": sortmock}
    ):
        assert pacman.group_info("testgroup")["default"] == ["pkg1", "pkg2"]


def test_group_diff():
    """
    Test if it shows the difference between installed and target group contents
    """

    listmock = MagicMock(return_value={"A": ["1.0"], "B": ["2.0"]})
    groupmock = MagicMock(
        return_value={
            "mandatory": [],
            "optional": [],
            "default": ["A", "C"],
            "conditional": [],
        }
    )
    with patch.dict(
        pacman.__salt__, {"pkg.list_pkgs": listmock, "pkg.group_info": groupmock}
    ):
        results = pacman.group_diff("testgroup")
        assert results["default"] == {"installed": ["A"], "not installed": ["C"]}



"""
Unit tests for salt.utils.templates.py
"""


import logging
import os
import sys
from collections import OrderedDict
from pathlib import PurePath, PurePosixPath

import pytest
import salt.utils.files
import salt.utils.templates
from tests.support.helpers import with_tempdir
from tests.support.mock import patch
from tests.support.unit import TestCase, skipIf

try:
    import Cheetah as _

    HAS_CHEETAH = True
except ImportError:
    HAS_CHEETAH = False

log = logging.getLogger(__name__)


class RenderTestCase(TestCase):
    def setUp(self):
        # Default context for salt.utils.templates.render_*_tmpl to work
        self.context = {
            "opts": {"cachedir": "/D", "__cli": "salt"},
            "saltenv": None,
        }

    ### Tests for Jinja (whitespace-friendly)
    def test_render_jinja_sanity(self):
        tmpl = """OK"""
        res = salt.utils.templates.render_jinja_tmpl(tmpl, dict(self.context))
        self.assertEqual(res, "OK")

    def test_render_jinja_evaluate(self):
        tmpl = """{{ "OK" }}"""
        res = salt.utils.templates.render_jinja_tmpl(tmpl, dict(self.context))
        self.assertEqual(res, "OK")

    def test_render_jinja_evaluate_multi(self):
        tmpl = """{% if 1 -%}OK{%- endif %}"""
        res = salt.utils.templates.render_jinja_tmpl(tmpl, dict(self.context))
        self.assertEqual(res, "OK")

    def test_render_jinja_variable(self):
        tmpl = """{{ var }}"""

        ctx = dict(self.context)
        ctx["var"] = "OK"
        res = salt.utils.templates.render_jinja_tmpl(tmpl, ctx)
        self.assertEqual(res, "OK")

    def test_render_jinja_tojson_sorted(self):
        templ = """thing: {{ var|tojson(sort_keys=True) }}"""
        expected = """thing: {"x": "xxx", "y": "yyy", "z": "zzz"}"""

        with patch.dict(self.context, {"var": {"z": "zzz", "y": "yyy", "x": "xxx"}}):
            res = salt.utils.templates.render_jinja_tmpl(templ, self.context)

        assert res == expected

    def test_render_jinja_tojson_unsorted(self):
        templ = """thing: {{ var|tojson(sort_keys=False) }}"""
        expected = """thing: {"z": "zzz", "x": "xxx", "y": "yyy"}"""

        # Values must be added to the dict in the expected order. This is
        # only necessary for older Pythons that don't remember dict order.
        d = OrderedDict()
        d["z"] = "zzz"
        d["x"] = "xxx"
        d["y"] = "yyy"

        with patch.dict(self.context, {"var": d}):
            res = salt.utils.templates.render_jinja_tmpl(templ, self.context)

        assert res == expected

    ### Tests for mako template
    def test_render_mako_sanity(self):
        tmpl = """OK"""
        res = salt.utils.templates.render_mako_tmpl(tmpl, dict(self.context))
        self.assertEqual(res, "OK")

    def test_render_mako_evaluate(self):
        tmpl = """${ "OK" }"""
        res = salt.utils.templates.render_mako_tmpl(tmpl, dict(self.context))
        self.assertEqual(res, "OK")

    def test_render_mako_evaluate_multi(self):
        tmpl = """
        % if 1:
        OK
        % endif
        """
        res = salt.utils.templates.render_mako_tmpl(tmpl, dict(self.context))
        stripped = res.strip()
        self.assertEqual(stripped, "OK")

    def test_render_mako_variable(self):
        tmpl = """${ var }"""

        ctx = dict(self.context)
        ctx["var"] = "OK"
        res = salt.utils.templates.render_mako_tmpl(tmpl, ctx)
        self.assertEqual(res, "OK")

    ### Tests for wempy template
    @skipIf(
        sys.version_info > (3,),
        "The wempy module is currently unsupported under Python3",
    )
    def test_render_wempy_sanity(self):
        tmpl = """OK"""
        res = salt.utils.templates.render_wempy_tmpl(tmpl, dict(self.context))
        self.assertEqual(res, "OK")

    @skipIf(
        sys.version_info > (3,),
        "The wempy module is currently unsupported under Python3",
    )
    def test_render_wempy_evaluate(self):
        tmpl = """{{="OK"}}"""
        res = salt.utils.templates.render_wempy_tmpl(tmpl, dict(self.context))
        self.assertEqual(res, "OK")

    @skipIf(
        sys.version_info > (3,),
        "The wempy module is currently unsupported under Python3",
    )
    def test_render_wempy_evaluate_multi(self):
        tmpl = """{{if 1:}}OK{{pass}}"""
        res = salt.utils.templates.render_wempy_tmpl(tmpl, dict(self.context))
        self.assertEqual(res, "OK")

    @skipIf(
        sys.version_info > (3,),
        "The wempy module is currently unsupported under Python3",
    )
    def test_render_wempy_variable(self):
        tmpl = """{{=var}}"""

        ctx = dict(self.context)
        ctx["var"] = "OK"
        res = salt.utils.templates.render_wempy_tmpl(tmpl, ctx)
        self.assertEqual(res, "OK")

    ### Tests for genshi template (xml-based)
    def test_render_genshi_sanity(self):
        tmpl = """<RU>OK</RU>"""
        res = salt.utils.templates.render_genshi_tmpl(tmpl, dict(self.context))
        self.assertEqual(res, "<RU>OK</RU>")

    def test_render_genshi_evaluate(self):
        tmpl = """<RU>${ "OK" }</RU>"""
        res = salt.utils.templates.render_genshi_tmpl(tmpl, dict(self.context))
        self.assertEqual(res, "<RU>OK</RU>")

    def test_render_genshi_evaluate_condition(self):
        tmpl = """<RU xmlns:py="http://genshi.edgewall.org/" py:if="1">OK</RU>"""
        res = salt.utils.templates.render_genshi_tmpl(tmpl, dict(self.context))
        self.assertEqual(res, "<RU>OK</RU>")

    def test_render_genshi_variable(self):
        tmpl = """<RU>$var</RU>"""

        ctx = dict(self.context)
        ctx["var"] = "OK"
        res = salt.utils.templates.render_genshi_tmpl(tmpl, ctx)
        self.assertEqual(res, "<RU>OK</RU>")

    def test_render_genshi_variable_replace(self):
        tmpl = """<RU xmlns:py="http://genshi.edgewall.org/" py:content="var">not ok</RU>"""

        ctx = dict(self.context)
        ctx["var"] = "OK"
        res = salt.utils.templates.render_genshi_tmpl(tmpl, ctx)
        self.assertEqual(res, "<RU>OK</RU>")

    ### Tests for cheetah template (line-oriented and xml-friendly)
    @skipIf(not HAS_CHEETAH, "The Cheetah Python module is missing.")
    def test_render_cheetah_sanity(self):
        tmpl = """OK"""
        res = salt.utils.templates.render_cheetah_tmpl(tmpl, dict(self.context))
        self.assertEqual(res, "OK")

    @skipIf(not HAS_CHEETAH, "The Cheetah Python module is missing.")
    def test_render_cheetah_evaluate(self):
        tmpl = """<%="OK"%>"""
        res = salt.utils.templates.render_cheetah_tmpl(tmpl, dict(self.context))
        self.assertEqual(res, "OK")

    @skipIf(not HAS_CHEETAH, "The Cheetah Python module is missing.")
    def test_render_cheetah_evaluate_xml(self):
        tmpl = """
        <% if 1: %>
        OK
        <% pass %>
        """
        res = salt.utils.templates.render_cheetah_tmpl(tmpl, dict(self.context))
        stripped = res.strip()
        self.assertEqual(stripped, "OK")

    @skipIf(not HAS_CHEETAH, "The Cheetah Python module is missing.")
    def test_render_cheetah_evaluate_text(self):
        tmpl = """
        #if 1
        OK
        #end if
        """

        res = salt.utils.templates.render_cheetah_tmpl(tmpl, dict(self.context))
        stripped = res.strip()
        self.assertEqual(stripped, "OK")

    @skipIf(not HAS_CHEETAH, "The Cheetah Python module is missing.")
    def test_render_cheetah_variable(self):
        tmpl = """$var"""

        ctx = dict(self.context)
        ctx["var"] = "OK"
        res = salt.utils.templates.render_cheetah_tmpl(tmpl, ctx)
        self.assertEqual(res.strip(), "OK")

    def test_render_jinja_cve_2021_25283(self):
        tmpl = """{{ [].__class__ }}"""
        ctx = dict(self.context)
        ctx["var"] = "OK"
        with pytest.raises(salt.exceptions.SaltRenderError):
            res = salt.utils.templates.render_jinja_tmpl(tmpl, ctx)


class MockRender:
    def __call__(self, tplstr, context, tmplpath=None):
        self.tplstr = tplstr
        self.context = context
        self.tmplpath = tmplpath
        return tplstr


class WrapRenderTestCase(TestCase):
    def assertDictContainsAll(self, actual, **expected):
        """Make sure dictionary contains at least all expected values"""
        actual = {key: actual[key] for key in expected if key in actual}
        self.assertEqual(expected, actual)

    def _test_generated_sls_context(self, tmplpath, sls, **expected):
        """Generic SLS Context Test"""
        # DeNormalize tmplpath
        tmplpath = str(PurePath(PurePosixPath(tmplpath)))
        if tmplpath.startswith("\\"):
            tmplpath = "C:{}".format(tmplpath)
        expected["tplpath"] = tmplpath
        actual = salt.utils.templates._generate_sls_context(tmplpath, sls)
        self.assertDictContainsAll(actual, **expected)

    @patch("salt.utils.templates.generate_sls_context")
    @with_tempdir()
    def test_sls_context_call(self, tempdir, generate_sls_context):
        """Check that generate_sls_context is called with proper parameters"""
        sls = "foo.bar"
        tmplpath = "/tmp/foo/bar.sls"

        slsfile = os.path.join(tempdir, "foo")
        with salt.utils.files.fopen(slsfile, "w") as fp:
            fp.write("{{ slspath }}")
        context = {"opts": {}, "saltenv": "base", "sls": sls}
        render = MockRender()
        wrapped = salt.utils.templates.wrap_tmpl_func(render)
        res = wrapped(slsfile, context=context, tmplpath=tmplpath)
        generate_sls_context.assert_called_with(tmplpath, sls)

    @patch("salt.utils.templates.generate_sls_context")
    @with_tempdir()
    def test_sls_context_no_call(self, tempdir, generate_sls_context):
        """Check that generate_sls_context is not called if sls is not set"""
        sls = "foo.bar"
        tmplpath = "/tmp/foo/bar.sls"

        slsfile = os.path.join(tempdir, "foo")
        with salt.utils.files.fopen(slsfile, "w") as fp:
            fp.write("{{ slspath }}")
        context = {"opts": {}, "saltenv": "base"}
        render = MockRender()
        wrapped = salt.utils.templates.wrap_tmpl_func(render)
        res = wrapped(slsfile, context=context, tmplpath=tmplpath)
        generate_sls_context.assert_not_called()

    def test_generate_sls_context__top_level(self):
        """generate_sls_context - top_level Use case"""
        self._test_generated_sls_context(
            "/tmp/boo.sls",
            "boo",
            tplfile="boo.sls",
            tpldir=".",
            tpldot="",
            slsdotpath="",
            slscolonpath="",
            sls_path="",
            slspath="",
        )

    def test_generate_sls_context__one_level_init_implicit(self):
        """generate_sls_context - Basic one level with implicit init.sls"""
        self._test_generated_sls_context(
            "/tmp/foo/init.sls",
            "foo",
            tplfile="foo/init.sls",
            tpldir="foo",
            tpldot="foo",
            slsdotpath="foo",
            slscolonpath="foo",
            sls_path="foo",
            slspath="foo",
        )

    def test_generate_sls_context__one_level_init_explicit(self):
        """generate_sls_context - Basic one level with explicit init.sls"""
        self._test_generated_sls_context(
            "/tmp/foo/init.sls",
            "foo.init",
            tplfile="foo/init.sls",
            tpldir="foo",
            tpldot="foo",
            slsdotpath="foo",
            slscolonpath="foo",
            sls_path="foo",
            slspath="foo",
        )

    def test_generate_sls_context__one_level(self):
        """generate_sls_context - Basic one level with name"""
        self._test_generated_sls_context(
            "/tmp/foo/boo.sls",
            "foo.boo",
            tplfile="foo/boo.sls",
            tpldir="foo",
            tpldot="foo",
            slsdotpath="foo",
            slscolonpath="foo",
            sls_path="foo",
            slspath="foo",
        )

    def test_generate_sls_context__one_level_repeating(self):
        """generate_sls_context - Basic one level with name same as dir

        (Issue #56410)
        """
        self._test_generated_sls_context(
            "/tmp/foo/foo.sls",
            "foo.foo",
            tplfile="foo/foo.sls",
            tpldir="foo",
            tpldot="foo",
            slsdotpath="foo",
            slscolonpath="foo",
            sls_path="foo",
            slspath="foo",
        )

    def test_generate_sls_context__two_level_init_implicit(self):
        """generate_sls_context - Basic two level with implicit init.sls"""
        self._test_generated_sls_context(
            "/tmp/foo/bar/init.sls",
            "foo.bar",
            tplfile="foo/bar/init.sls",
            tpldir="foo/bar",
            tpldot="foo.bar",
            slsdotpath="foo.bar",
            slscolonpath="foo:bar",
            sls_path="foo_bar",
            slspath="foo/bar",
        )

    def test_generate_sls_context__two_level_init_explicit(self):
        """generate_sls_context - Basic two level with explicit init.sls"""
        self._test_generated_sls_context(
            "/tmp/foo/bar/init.sls",
            "foo.bar.init",
            tplfile="foo/bar/init.sls",
            tpldir="foo/bar",
            tpldot="foo.bar",
            slsdotpath="foo.bar",
            slscolonpath="foo:bar",
            sls_path="foo_bar",
            slspath="foo/bar",
        )

    def test_generate_sls_context__two_level(self):
        """generate_sls_context - Basic two level with name"""
        self._test_generated_sls_context(
            "/tmp/foo/bar/boo.sls",
            "foo.bar.boo",
            tplfile="foo/bar/boo.sls",
            tpldir="foo/bar",
            tpldot="foo.bar",
            slsdotpath="foo.bar",
            slscolonpath="foo:bar",
            sls_path="foo_bar",
            slspath="foo/bar",
        )

    def test_generate_sls_context__two_level_repeating(self):
        """generate_sls_context - Basic two level with name same as dir

        (Issue #56410)
        """
        self._test_generated_sls_context(
            "/tmp/foo/foo/foo.sls",
            "foo.foo.foo",
            tplfile="foo/foo/foo.sls",
            tpldir="foo/foo",
            tpldot="foo.foo",
            slsdotpath="foo.foo",
            slscolonpath="foo:foo",
            sls_path="foo_foo",
            slspath="foo/foo",
        )

    @patch("salt.utils.templates._generate_sls_context_legacy", return_value="legacy")
    @patch("salt.utils.templates._generate_sls_context", return_value="new")
    @patch("salt.utils.templates.features.get", return_value=True)
    def test_feature_flag_on(self, feature_get, new_impl, legacy_impl):
        """Test feature flag selection with FF on"""
        tplpath = "tplpath"
        sls = "sls"
        self.assertEqual("new", salt.utils.templates.generate_sls_context(tplpath, sls))
        new_impl.assert_called_with(tplpath, sls)
        legacy_impl.assert_not_called()

    @patch("salt.utils.templates._generate_sls_context_legacy", return_value="legacy")
    @patch("salt.utils.templates._generate_sls_context", return_value="new")
    @patch("salt.utils.templates.features.get", return_value=False)
    def test_feature_flag_off(self, feature_get, new_impl, legacy_impl):
        """Test feature flag selection with FF on"""
        tplpath = "tplpath"
        sls = "sls"
        self.assertEqual(
            "legacy", salt.utils.templates.generate_sls_context(tplpath, sls)
        )

        new_impl.assert_not_called()
        legacy_impl.assert_called_with(tplpath, sls)

    @skipIf(sys.platform == "win32", "Backslash not possible under windows")
    def test_generate_sls_context__backslash_in_path(self):
        """generate_sls_context - Handle backslash in path on non-windows"""
        self._test_generated_sls_context(
            "/tmp/foo/foo\\foo.sls",
            "foo.foo\\foo",
            tplfile="foo/foo\\foo.sls",
            tpldir="foo",
            tpldot="foo",
            slsdotpath="foo",
            slscolonpath="foo",
            sls_path="foo",
            slspath="foo",
        )
